import asyncio
import re

from imbi.endpoints import (base, cookie_cutters, environments, fact_types,
                            groups, namespaces, project_link_types,
                            project_types)
from imbi.opensearch import project

CONFIGURATION_TYPE_SQL = re.sub(
    r'\s+', ' ', """\
        SELECT unnest(enum_range(null, null::v1.configuration_type))
            AS configuration_type""")


class RequestHandler(base.RequestHandler):
    NAME = 'metadata'

    async def get(self) -> None:
        """Return all metadata in a single request"""
        project_index = project.ProjectIndex(self.application)
        results = await asyncio.gather(
            self.postgres_execute(
                cookie_cutters.CollectionRequestHandler.COLLECTION_SQL,
                metric_name='cookie-cutters'),
            self.postgres_execute(
                environments.CollectionRequestHandler.COLLECTION_SQL,
                metric_name='environments'),
            self.postgres_execute(
                groups.CollectionRequestHandler.COLLECTION_SQL,
                metric_name='groups'),
            self.postgres_execute(
                namespaces.CollectionRequestHandler.COLLECTION_SQL,
                metric_name='namespaces'),
            self.postgres_execute(CONFIGURATION_TYPE_SQL,
                                  metric_name='configuration-types'),
            self.postgres_execute(
                fact_types.CollectionRequestHandler.COLLECTION_SQL,
                metric_name='project-fact-types'),
            self.postgres_execute(
                project_link_types.CollectionRequestHandler.COLLECTION_SQL,
                metric_name='project-link-types'),
            self.postgres_execute(
                project_types.CollectionRequestHandler.COLLECTION_SQL,
                metric_name='project-types'),
            project_index.searchable_fields(),
            self.postgres_execute(
                'SELECT name, authorization_endpoint, client_id, callback_url'
                '  FROM v1.oauth2_integrations',
                metric_name='integrations'),
        )

        automations = self.settings['automations']

        oauth_details = results[9].rows
        gitlab = {'enabled': False}
        cfg = automations.get('gitlab')
        if cfg:
            gitlab['project_link_type_id'] = cfg.get('project_link_type_id')
            gitlab_auth = [r for r in oauth_details if r['name'] == 'gitlab']
            if gitlab_auth:
                gitlab_auth = gitlab_auth[0]
                gitlab.update({
                    'authorizationEndpoint': gitlab_auth[
                        'authorization_endpoint'],
                    'clientId': gitlab_auth['client_id'],
                    'redirectURI': gitlab_auth['callback_url'],
                })
            gitlab['enabled'] = (cfg['enabled']
                                 and gitlab.get('clientId') is not None)

        github = {'enabled': False}
        cfg = automations.get('github')
        if cfg:
            github['project_link_type_id'] = cfg.get('project_link_type_id')
            github_auth = [r for r in oauth_details if r['name'] == 'github']
            if github_auth:
                github_auth = github_auth[0]
                github.update({
                    'authorizationEndpoint': github_auth[
                        'authorization_endpoint'],
                    'clientId': github_auth['client_id'],
                    'redirectURI': github_auth['callback_url'],
                })
            github['enabled'] = (cfg['enabled']
                                 and github.get('clientId') is not None)

        sentry = {'enabled': False}
        cfg = automations.get('sentry')
        if cfg:
            sentry['project_link_type_id'] = cfg.get('project_link_type_id')
            sentry['enabled'] = cfg['enabled']

        self.send_response({
            'integrations': {
                'grafana': {
                    'enabled': automations['grafana']['enabled'],
                    'project_link_type_id': automations['grafana']
                    ['project_link_type_id']
                },
                'github': github,
                'gitlab': gitlab,
                'sentry': {
                    'enabled': automations['sentry']['enabled'],
                    'project_link_type_id': automations['sentry']
                    ['project_link_type_id']
                },
                'sonarqube': {
                    'enabled': automations['sonarqube']['enabled'],
                    'project_link_type_id': automations['sonarqube']
                    ['project_link_type_id']
                }
            },
            'metadata': {
                'cookie_cutters': results[0].rows,
                'environments': results[1].rows,
                'groups': results[2].rows,
                'namespaces': results[3].rows,
                'project_configuration_types': results[4].rows,
                'project_fact_types': results[5].rows,
                'project_link_types': results[6].rows,
                'project_types': results[7].rows
            },
            'opensearch': {
                'fields': results[8]
            },
            'ops_log_ticket_slug_template': self.
            settings['ops_log_ticket_slug_template'],
            'project_url_template': self.settings['project_url_template'],
            'ssm_prefix_template': self.application.
            settings['project_configuration']['ssm_prefix_template']
        })



from __future__ import annotations

import collections
import http
import typing

import sprockets_postgres
from psycopg2 import sql

import imbi.models
import imbi.opensearch.project
from imbi import errors
from imbi.endpoints import base
from imbi.endpoints.project_sbom import graph, models
from imbi.endpoints.components import scoring


class SBOMInjectionHandler(base.PydanticHandlerMixin,
                           base.AuthenticatedRequestHandler):
    NAME = 'sbom-injection'

    async def put(self, project_id: str) -> None:
        try:
            project_id = int(project_id)
        except ValueError:  # pragma: no-cover -- URL pattern prevents this
            raise errors.ItemNotFound

        sbom: models.SBOM = self.parse_request_body_as(models.SBOM)
        target_ref = await self._find_project_ref(sbom)
        if target_ref not in {d.ref for d in sbom.dependencies}:
            raise errors.UnprocessableEntity(
                'target_ref %r is not in the SBoM dependency list', target_ref)

        components: dict[models.BOMRef, models.Component] = {}
        package_urls: set[models.PackageURL] = set()
        for component in sbom.components:
            components[component.bom_ref] = component
            if component.package_purl:
                package_urls.add(component.package_purl)

        project = await imbi.models.project(project_id, self.application)
        if not project:
            raise errors.ItemNotFound

        async with self.application.postgres_connector(
                on_error=self.on_postgres_error,
                on_duration=self.on_postgres_timing,
        ) as connector:
            connector: sprockets_postgres.PostgresConnector
            project_components: set[tuple[str, int]] = set()

            dependency_graph = graph.DependencyGraph(sbom.dependencies)
            version_map = ComponentVersionMap(connector, package_urls)
            for bom_ref in dependency_graph.all_dependencies(target_ref):
                try:
                    component = components[bom_ref]
                except KeyError:
                    raise errors.UnprocessableEntity(
                        'unknown component %r referenced as dependency of %r',
                        bom_ref, target_ref)
                if not component.purl:
                    self.logger.warning(
                        'skipping dependency %r, no package URL', bom_ref)
                elif not component.version:
                    self.logger.warning('skipping dependency %r, no version',
                                        bom_ref)
                else:
                    version_id = await version_map.get_version_id(component)
                    project_components.add(
                        (component.package_purl, version_id))
                    # update the project model instance for indexing purposes
                    # see _reindex_project() and imbi.models.project()
                    project.components.append({
                        'name': component.package_purl,
                        'version': component.version
                    })

            await self._insert_project_components(connector, project,
                                                  project_components)
            self.logger.debug('processed %s components for project %s(%s)',
                              len(project_components), project.slug,
                              project.id)

            await scoring.update_component_score_for_project(
                project.id, connector, self.application, index_project=False)

        await self._reindex_project(project)

        self.set_status(http.HTTPStatus.NO_CONTENT)

    async def _find_project_ref(self, sbom):
        target_ref: models.BOMRef | None = None
        if qargs := self.get_query_arguments('target_ref'):
            target_ref = qargs[0]
            self.logger.debug('using target_ref from query: %r', target_ref)
            if len(qargs) > 1:
                self.logger.warning(
                    'received %s target_ref query parameters, using %s',
                    len(qargs), target_ref)
        if target_ref is None and sbom.metadata.component:
            target_ref = sbom.metadata.component.bom_ref
        if target_ref is None:
            raise errors.BadRequest(
                'failed to find SBoM reference for project',
                detail=("Project's identifier cannot be determined from the "
                        'SBoM, you need to explicitly include it with the '
                        'target_ref query parameter'),
            )
        return target_ref

    async def _insert_project_components(
        self,
        connector: sprockets_postgres.PostgresConnector,
        project: imbi.models.Project,
        component_versions: typing.Iterable[tuple[str, int]],
        *,
        batch_size: int = 250,
    ) -> None:
        async with connector.transaction() as transaction:
            transaction: sprockets_postgres.PostgresConnector
            result = await transaction.execute(
                'DELETE FROM v1.project_components'
                ' WHERE project_id = %(project_id)s',
                {'project_id': project.id},
                metric_name='delete-project-components')
            self.logger.debug('removed %s components for project %s',
                              result.row_count, project.id)

            rows_to_add = [(project.id, package_url, version_id)
                           for package_url, version_id in component_versions]
            while rows_to_add:
                values = sql.SQL(',').join(
                    sql.Literal(r) for r in rows_to_add[:batch_size])
                query = sql.SQL(
                    'INSERT INTO v1.project_components ('
                    '            project_id, package_url, version_id)'
                    '     VALUES {}').format(values)
                await transaction.execute(
                    query.as_string(transaction.cursor.raw),
                    metric_name='insert-project-component-batch')
                del rows_to_add[:batch_size]

    async def _reindex_project(self, project: imbi.models.Project) -> None:
        index = imbi.opensearch.project.ProjectIndex(self.application)
        await index.index_document(project)


class ComponentVersionMap:
    def __init__(self, connector: sprockets_postgres.PostgresConnector,
                 package_urls: set[models.PackageURL]) -> None:
        self._purls_from_bom = package_urls.copy()
        self._cache: dict[models.PackageURL, models.VersionMap] = {}
        self._cache_primed = False
        self._connector = connector

    async def get_version_id(self, component: models.Component) -> int:
        if not self._cache_primed:
            await self._fetch_components()
        try:
            return self._cache[component.package_purl][component.version]
        except KeyError:
            return await self._insert_component_version(component)

    async def _fetch_components(self) -> None:
        """Fetch component versions from the database for PURLs in the BoM"""
        cache = collections.defaultdict(dict)
        result = await self._connector.execute(
            'SELECT id, package_url, version'
            '  FROM v1.component_versions'
            ' WHERE package_url = ANY(%s)', [list(self._purls_from_bom)],
            metric_name='get-component-versions')
        for row in result:
            cache[row['package_url']][row['version']] = row['id']
        self._cache.clear()
        self._cache.update(cache)
        self._cache_primed = True

    async def _insert_component_version(self,
                                        component: models.Component) -> int:
        try:
            component_versions = self._cache[component.package_purl]
        except KeyError:
            await self._connector.execute(
                'INSERT INTO v1.components(package_url, name, home_page,'
                '                          created_by)'
                '     VALUES (%(package_url)s, %(name)s, %(home_page)s,'
                '             %(created_by)s) '
                'ON CONFLICT (package_url) DO NOTHING', {
                    'package_url': component.package_purl,
                    'name': component.name,
                    'home_page': component.home_page,
                    'created_by': 'system',
                },
                metric_name='insert-component')
            component_versions = self._cache[component.package_purl] = {}

        try:
            return component_versions[component.version]
        except KeyError:
            result = await self._connector.execute(
                'INSERT INTO v1.component_versions(package_url, version)'
                '     VALUES (%(package_url)s, %(version)s) '
                'ON CONFLICT (package_url, version)'
                '  DO UPDATE SET version = %(version)s'
                '  RETURNING id', {
                    'package_url': component.package_purl,
                    'version': component.version,
                },
                metric_name='insert-component-version')
            component_versions[component.version] = result.row['id']
            return result.row['id']



from __future__ import annotations

import copy
import unittest.mock
import uuid

import yarl
from tornado import httputil, web

from imbi import cors


class OriginTests(unittest.TestCase):
    def test_defaults(self):
        origins = cors.Origins()
        self.assertTrue(origins.allow_any)
        self.assertIn(f'https://{uuid.uuid4()}.example.com', origins)

    def test_adding_origin_disables_allow_any(self):
        origins = cors.Origins()
        self.assertTrue(origins.allow_any)

        origins.add('https://example.com/')
        self.assertFalse(origins.allow_any)

    def test_adding_relative_origin(self):
        origins = cors.Origins()
        with self.assertRaises(ValueError):
            origins.add('/foo')


class ConfigTests(unittest.TestCase):
    def test_default_configuration(self):
        cfg = cors.CORSConfig()
        self.assertTrue(cfg.allow_credentials)
        self.assertSetEqual(set(), cfg.allowed_methods)
        self.assertIn('https://example.com', cfg.allowed_origins)
        self.assertIn('Cache-Control', cfg.exposed_headers)
        self.assertIn('Date', cfg.exposed_headers)
        self.assertIn('Last-Modified', cfg.exposed_headers)
        self.assertIn('Link', cfg.exposed_headers)

    def test_that_allow_credentials_can_be_set(self):
        cfg = cors.CORSConfig(allow_credentials=False)
        self.assertFalse(cfg.allow_credentials)

    def test_that_allow_any_origin_can_be_set(self):
        cfg = cors.CORSConfig(allow_any_origin=False)
        self.assertNotIn('https://example.com', cfg.allowed_origins)

    def test_that_allow_methods_can_be_set(self):
        cfg = cors.CORSConfig(allow_methods={'POST'})
        self.assertSetEqual({'POST'}, cfg.allowed_methods)

    def test_updating(self):
        cfg = cors.CORSConfig(allow_any_origin=True,
                              allow_credentials=True,
                              allow_methods={'POST'})

        cfg.update(allow_any_origin=False, allow_credentials=False)
        self.assertFalse(cfg.allowed_origins.allow_any)
        self.assertFalse(cfg.allow_credentials)
        self.assertSetEqual({'POST'}, cfg.allowed_methods)

        # no-op is allowed
        cfg.update()
        self.assertFalse(cfg.allowed_origins.allow_any)
        self.assertFalse(cfg.allow_credentials)
        self.assertSetEqual({'POST'}, cfg.allowed_methods)

        # allow_methods appends
        cfg.update(allow_methods={'DELETE'})
        self.assertFalse(cfg.allowed_origins.allow_any)
        self.assertFalse(cfg.allow_credentials)
        self.assertSetEqual({'DELETE', 'POST'}, cfg.allowed_methods)

        # allow_origins appends
        cfg.update(allow_origins={'https://example.net'})
        self.assertSetEqual({yarl.URL('https://example.net')},
                            cfg.allowed_origins._origins)

        # exposed_headers appends
        starting = cfg.exposed_headers.copy()
        cfg.update(exposed_headers={'ETag'})
        self.assertSetEqual(starting.union({'ETag'}), cfg.exposed_headers)

        # max_age overwrites
        cfg.update(max_age=0)
        self.assertEqual(0, cfg.max_age)

    def test_copying(self):
        source = cors.CORSConfig(allow_methods={'DELETE', 'GET'})
        source.allowed_origins.add('https://example.com')

        clone = copy.copy(source)
        self.assertEqual(source.allow_credentials, clone.allow_credentials)
        self.assertSetEqual(source.allowed_methods, clone.allowed_methods)


class RequestProcessingTestCase(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.request = httputil.HTTPServerRequest(
            headers=httputil.HTTPHeaders({
                'Origin': 'https://example.com',
            }),
            connection=unittest.mock.Mock(),
            uri='/')
        self.response_headers = httputil.HTTPHeaders()
        self.handler = web.RequestHandler(web.Application(), self.request)
        self.handler.set_header = self.response_headers.__setitem__
        self.handler.add_header = self.response_headers.add


class PreflightProcessingTests(RequestProcessingTestCase):
    def setUp(self) -> None:
        super().setUp()
        self.cors_processor = cors.CORSProcessor(
            cors.CORSConfig(allow_methods={'POST'}))
        self.request.method = 'OPTIONS'
        self.request.headers['Access-Control-Request-Method'] = 'POST'

    def assert_preflight_failure(self):
        self.assertFalse(self.cors_processor.ok)
        self.assertIn('Origin', self.response_headers.get_list('Vary'))
        self.assertNotIn('Access-Control-Allow-Credentials',
                         self.response_headers)

    def test_correct_cors_preflight(self):
        self.cors_processor.process_request(self.handler)
        self.assertTrue(self.cors_processor.ok)
        self.assertEqual(
            self.request.headers['Origin'],
            self.response_headers.get('Access-Control-Allow-Origin'))
        self.assertEqual(
            'POST', self.response_headers.get('Access-Control-Allow-Methods'))
        self.assertNotIn('Access-Control-Allow-Headers', self.response_headers)
        self.assertEqual(str(self.cors_processor.config.max_age),
                         self.response_headers.get('Access-Control-Max-Age'))
        self.assertIn('Origin', self.response_headers.get_list('Vary'))

    def test_correct_cors_preflight_with_request_headers(self):
        self.request.headers.add('Access-Control-Request-Headers',
                                 'Cache-Control')
        self.cors_processor.process_request(self.handler)
        self.assertTrue(self.cors_processor.ok)
        self.assertIn(
            'Cache-Control',
            self.response_headers.get_list('Access-Control-Allow-Headers'))

    def test_correct_cors_preflight_with_exposed_headers(self):
        self.cors_processor.config.exposed_headers.add('Link')
        self.cors_processor.process_request(self.handler)
        self.assertTrue(self.cors_processor.ok)
        self.assertIn(
            'Link',
            self.response_headers.get_list('Access-Control-Expose-Headers'))

    def test_preflight_without_ac_request_method(self):
        del self.request.headers['Access-Control-Request-Method']
        self.cors_processor.process_request(self.handler)
        self.assert_preflight_failure()
        self.assertEqual('*',
                         self.response_headers['Access-Control-Allow-Origin'])

    def test_preflight_without_origin(self):
        del self.request.headers['Origin']
        self.cors_processor.process_request(self.handler)
        self.assert_preflight_failure()
        self.assertEqual('*',
                         self.response_headers['Access-Control-Allow-Origin'])

    def test_preflight_with_explicit_allowed_origin(self):
        self.cors_processor.config.allowed_origins.allow_any = False
        self.cors_processor.config.allowed_origins.add(
            self.request.headers['Origin'])
        self.cors_processor.process_request(self.handler)
        self.assertTrue(self.cors_processor.ok)
        self.assertEqual(self.request.headers['Origin'],
                         self.response_headers['Access-Control-Allow-Origin'])

    def test_preflight_without_explicit_allowed_origin(self):
        self.cors_processor.config.allowed_origins.allow_any = False
        self.request.headers['Origin'] = 'https://example.net'
        self.cors_processor.process_request(self.handler)
        self.assert_preflight_failure()
        self.assertNotIn('Access-Control-Allow-Origin', self.response_headers)


class RequestProcessingTests(RequestProcessingTestCase):
    def setUp(self) -> None:
        super().setUp()
        self.request.method = 'DELETE'

    def process_request(self, **overrides):
        self.response_headers.clear()
        cors_processor = cors.CORSProcessor(
            cors.CORSConfig(allow_methods={'DELETE'}), **overrides)
        cors_processor.process_request(self.handler)

    def test_allowed_request(self):
        self.process_request()
        self.assertEqual(self.request.headers['Origin'],
                         self.response_headers['Access-Control-Allow-Origin'])
        self.assertEqual(
            'true', self.response_headers['Access-Control-Allow-Credentials'])

    def test_request_with_unallowed_method(self):
        # NB -- CORS does not require that the request method matches
        #       the preflight Access-Control-Request-Method    o_O
        self.request.method = 'POST'
        self.process_request()
        self.assertEqual(self.request.headers['Origin'],
                         self.response_headers['Access-Control-Allow-Origin'])
        self.assertEqual(
            'true', self.response_headers['Access-Control-Allow-Credentials'])

    def test_request_without_origin(self):
        del self.request.headers['Origin']
        self.process_request()
        self.assertEqual('*',
                         self.response_headers['Access-Control-Allow-Origin'])
        self.assertNotIn('Access-Control-Allow-Credentials',
                         self.response_headers)

        self.process_request(allow_any_origin=False)
        self.assertNotIn('Access-Control-Allow-Origin', self.response_headers)
        self.assertNotIn('Access-Control-Allow-Credentials',
                         self.response_headers)

    def test_that_allowed_origin_is_widened_without_credentials(self):
        self.process_request(allow_any_origin=True, allow_credentials=False)
        self.assertEqual('*',
                         self.response_headers['Access-Control-Allow-Origin'])

        self.process_request(allow_credentials=False,
                             allow_origins={self.request.headers['Origin']})
        self.assertEqual(self.request.headers['Origin'],
                         self.response_headers['Access-Control-Allow-Origin'])


class MixinTests(unittest.IsolatedAsyncioTestCase):
    class Handler(cors.CORSMixin):
        def delete(self) -> None:
            self.set_status(204)

    def setUp(self) -> None:
        super().setUp()
        self.origin = 'https://example.com'
        self.application = web.Application()

    async def run_handler(self,
                          method: str,
                          *,
                          headers: None | dict = None,
                          **cors_overrides) -> httputil.HTTPHeaders:
        self.Handler.cors_overrides.clear()
        self.Handler.cors_overrides.update(cors_overrides)

        request = httputil.HTTPServerRequest(method=method,
                                             headers=httputil.HTTPHeaders(
                                                 {'Origin': self.origin}),
                                             connection=unittest.mock.Mock(),
                                             uri='/')
        if headers:
            for name, value in headers.items():
                request.headers[name] = value

        handler = self.Handler(self.application, request)
        response_headers = httputil.HTTPHeaders()
        handler.set_header = response_headers.__setitem__
        handler.add_header = response_headers.add
        await handler._execute([])

        return response_headers

    async def test_preflight_request_without_configuration(self):
        response_headers = await self.run_handler(
            'OPTIONS', headers={'Access-Control-Request-Method': 'DELETE'})
        self.assertEqual(self.origin,
                         response_headers['Access-Control-Allow-Origin'])
        self.assertCountEqual(
            ['DELETE'],
            response_headers.get_list('Access-Control-Allow-Methods'))

    async def test_preflight_request_with_methods_defined(self):
        response_headers = await self.run_handler(
            'OPTIONS',
            headers={'Access-Control-Request-Method': 'GET'},
            allow_methods={'GET'})
        self.assertEqual(self.origin,
                         response_headers['Access-Control-Allow-Origin'])
        self.assertCountEqual(
            ['GET'], response_headers.get_list('Access-Control-Allow-Methods'))

    async def test_preflight_request_with_application_config(self):
        self.application.cors_config = cors.CORSConfig(allow_any_origin=False,
                                                       allow_credentials=True,
                                                       max_age=300)
        self.application.cors_config.allowed_origins.add(self.origin)
        response_headers = await self.run_handler(
            'OPTIONS', headers={'Access-Control-Request-Method': 'DELETE'})
        self.assertEqual(self.origin,
                         response_headers['Access-Control-Allow-Origin'])
        self.assertCountEqual(
            ['DELETE'],
            response_headers.get_list('Access-Control-Allow-Methods'))
        self.assertEqual(str(self.application.cors_config.max_age),
                         response_headers['Access-Control-Max-Age'])

    async def test_non_preflight_response_headers(self):
        response_headers = await self.run_handler('DELETE')
        self.assertIn('Origin', response_headers.get('Vary'))
        self.assertEqual(self.origin,
                         response_headers['Access-Control-Allow-Origin'])

    async def test_cors_headers_on_request_failure(self):
        with unittest.mock.patch.object(
                self.Handler, 'delete',
                new_callable=unittest.mock.MagicMock) as delete:
            delete.__name__ = 'delete'
            delete.side_effect = RuntimeError('injected failure')
            response_headers = await self.run_handler('DELETE')

        self.assertIn('Origin', response_headers.get('Vary'))
        self.assertEqual(self.origin,
                         response_headers['Access-Control-Allow-Origin'])



from __future__ import annotations

import enum
import typing_extensions as typing

import pydantic

from imbi import semver, errors
from imbi.endpoints import base


class ComponentStatus(str, enum.Enum):
    ACTIVE = 'Active'
    DEPRECATED = 'Deprecated'
    FORBIDDEN = 'Forbidden'


class ProjectComponentStatus(str, enum.Enum):
    UNSCORED = 'Unscored'
    UP_TO_DATE = 'Up-to-date'
    OUTDATED = 'Outdated'
    DEPRECATED = 'Deprecated'
    FORBIDDEN = 'Forbidden'


# these are considered "out of date" for scoring purposes
OUTDATED_COMPONENT_STATUSES = frozenset((
    ProjectComponentStatus.DEPRECATED.value,
    ProjectComponentStatus.FORBIDDEN.value,
    ProjectComponentStatus.OUTDATED.value,
))

# Describes the expected `stats` dict for ProjectStatus.calculate
ProjectStatusCalculateStats = typing.TypedDict(
    'ProjectStatusCalculateStats',
    {e.value: int
     for e in ProjectComponentStatus},
    total=False,
)


class ProjectStatus(int, enum.Enum):
    """Component Score project fact values"""
    OKAY = 100
    NEEDS_WORK = 80
    UNACCEPTABLE = 20

    @classmethod
    def calculate(cls, stats: ProjectStatusCalculateStats) -> typing.Self:
        deprecated = stats[ProjectComponentStatus.DEPRECATED.value]
        forbidden = stats[ProjectComponentStatus.FORBIDDEN.value]
        outdated = stats[ProjectComponentStatus.OUTDATED.value]
        up_to_date = stats[ProjectComponentStatus.UP_TO_DATE.value]
        if not deprecated and not forbidden and not outdated:
            return cls.OKAY
        if not forbidden and (up_to_date > outdated or deprecated):
            return cls.NEEDS_WORK
        return cls.UNACCEPTABLE


class Component(pydantic.BaseModel):
    package_url: str = pydantic.constr(pattern=r'^pkg:')
    name: str
    status: ComponentStatus
    icon_class: str
    active_version: typing.Union[semver.VersionRange, None]
    home_page: typing.Union[str, None]


class ComponentToken(base.PaginationToken):
    """Pagination token that includes the starting package URL"""
    def __init__(self, *, starting_package: str = '', **kwargs) -> None:
        super().__init__(starting_package=starting_package, **kwargs)

    def with_first(self, value: dict[str, object]) -> typing.Self:
        kwargs = self.as_dict(starting_package=value['package_url'])
        return ComponentToken(**kwargs)


class ProjectComponentsToken(base.PaginationToken):
    """Pagination token that includes the starting package URL and project"""
    def __init__(self,
                 *,
                 starting_package: str = '',
                 project_id: int | str,
                 **kwargs) -> None:
        try:
            project_id = int(project_id)
        except ValueError:
            raise errors.BadRequest('Invalid project id %r, expected integer',
                                    project_id)
        super().__init__(starting_package=starting_package,
                         project_id=project_id,
                         **kwargs)

    def with_first(self, value: dict[str, object]) -> typing.Self:
        kwargs = self.as_dict(starting_package=value['package_url'])
        return ProjectComponentsToken(**kwargs)



"""
Request Handler used for rendering errors

"""
from tornado import web

from imbi.endpoints import base


class RequestHandler(base.RequestHandler):

    NAME = 'not-found'

    def get(self, *args, **kwargs):
        raise web.HTTPError(404)



import collections
import re
import typing
import urllib.parse

import aioboto3
import aioredis
import botocore.exceptions
import jsonpatch
import sprockets.mixins.http
from tornado import web

from imbi import errors, oauth2, user
from imbi.clients import aws
from imbi.endpoints import base
from imbi.endpoints.ssm import models


def path_prefix(template: str, project_slug: str, project_type_slug: str,
                namespace_slug: str) -> str:
    return template.replace('{project_slug}', project_slug).replace(
        '{project_type_slug}',
        project_type_slug).replace('{namespace_slug}', namespace_slug)


class SSMRequestHandler(sprockets.mixins.http.HTTPClientMixin,
                        base.ValidatingRequestHandler):
    GET_PROJECT_INFO_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT n.id AS namespace_id,
               n.aws_ssm_slug AS namespace_slug,
               p.slug AS project_slug,
               t.slug AS project_type_slug,
               p.configuration_type,
               p.environments
          FROM v1.namespaces AS n
          JOIN v1.projects AS p
            ON p.namespace_id = n.id
          JOIN v1.project_types AS t
            ON p.project_type_id = t.id
         WHERE p.id = %(project_id)s""")

    GET_ROLE_ARN_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT role_arn
          FROM v1.aws_roles
         WHERE environment = %(environment)s
           AND namespace_id = %(namespace_id)s""")

    @property
    def _redis(self) -> aioredis.Redis:
        return self.application.session_redis

    async def get_parameter(self, environment, namespace_id, aws_session,
                            user_id_token, user_refresh_token, name) -> str:
        role_arn = await self.get_role_arn(environment, namespace_id)
        if not role_arn:
            raise errors.Forbidden('No role ARN found for namespace %d in %s',
                                   namespace_id, environment)
        creds = await self.get_aws_credentials(aws_session, role_arn,
                                               self.current_user,
                                               user_id_token,
                                               user_refresh_token)

        try:
            param = await aws.get_parameter(name, aws_session,
                                            creds['access_key_id'],
                                            creds['secret_access_key'],
                                            creds['session_token'])
        except botocore.exceptions.ClientError as error:
            code = error.response['Error']['Code']
            if code == 'ParameterNotFound':
                raise errors.BadRequest('ParameterNotFound')
            else:
                raise errors.InternalServerError(code)
        return param['Value']

    async def delete_param(self, environment, namespace_id, aws_session,
                           user_id_token, user_refresh_token, name):
        role_arn = await self.get_role_arn(environment, namespace_id)
        if not role_arn:
            raise errors.Forbidden('No role ARN found for namespace %d in %s',
                                   namespace_id, environment)
        creds = await self.get_aws_credentials(aws_session, role_arn,
                                               self.current_user,
                                               user_id_token,
                                               user_refresh_token)

        try:
            await aws.delete_parameter(aws_session, creds['access_key_id'],
                                       creds['secret_access_key'],
                                       creds['session_token'], name)
        except botocore.exceptions.ClientError as error:
            code = error.response['Error']['Code']
            if code == 'ParameterNotFound':
                return
            else:
                raise errors.InternalServerError(code)

    async def put_parameter(self,
                            environment,
                            namespace_id,
                            aws_session,
                            user_id_token,
                            user_refresh_token,
                            name,
                            value,
                            param_type=None,
                            overwrite=False):
        role_arn = await self.get_role_arn(environment, namespace_id)
        if not role_arn:
            raise errors.Forbidden('No role ARN found for namespace %d in %s',
                                   namespace_id, environment)

        creds = await self.get_aws_credentials(aws_session, role_arn,
                                               self.current_user,
                                               user_id_token,
                                               user_refresh_token)
        try:
            await aws.put_parameter(aws_session, creds['access_key_id'],
                                    creds['secret_access_key'],
                                    creds['session_token'], name, value,
                                    param_type, overwrite)
        except botocore.exceptions.ClientError as error:
            code = error.response['Error']['Code']
            if code == 'UnsupportedParameterType':
                raise errors.BadRequest('UnsupportedParameterType')
            elif code == 'ParameterAlreadyExists':
                raise errors.BadRequest('ParameterAlreadyExists')
            elif code == 'ParameterLimitExceeded':
                raise errors.BadRequest('ParameterLimitExceeded')
            else:
                raise errors.InternalServerError(code)

    async def get_role_arn(self, environment,
                           namespace_id) -> typing.Optional[str]:
        result = await self.postgres_execute(self.GET_ROLE_ARN_SQL, {
            'environment': environment,
            'namespace_id': namespace_id
        }, 'get-role-arn')
        if not result.row:
            return None
        return result.row['role_arn']

    async def _get_project_info(self, project_id) -> dict:
        result = await self.postgres_execute(self.GET_PROJECT_INFO_SQL,
                                             {'project_id': project_id},
                                             'get-ssm-project-info')
        project_info = result.row

        if project_info['configuration_type'] != 'ssm':
            raise errors.ItemNotFound(
                'Project %s does not use SSM configuration',
                project_info['project_slug'],
                title='Wrong configuration type for project')

        return project_info

    async def _get_google_tokens(self) -> oauth2.IntegrationToken:
        if not self.settings['google']['enabled']:
            raise errors.Forbidden('Google integration is disabled',
                                   title='Google integration is disabled')

        if not self.settings['google']['integration_name']:
            raise errors.Forbidden(
                'No Google integration specified in configuration',
                title='No Google integration specified in configuration')

        tokens = await self.current_user.fetch_integration_tokens(
            self.settings['google']['integration_name'])
        if len(tokens) == 0:
            raise errors.Forbidden(
                'No OAuth 2.0 tokens for %s',
                self.current_user.username,
                title=('Not authorized to access SSM in this namespace &'
                       ' environment'))
        return tokens[0]

    async def get_aws_credentials(self,
                                  aws_session: aioboto3.Session,
                                  role_arn,
                                  imbi_user: user.User,
                                  id_token,
                                  refresh_token,
                                  is_retry=False) -> dict:
        """Return (potentially cached or refreshed) AWS credentials."""
        key = f'{imbi_user.username}:{role_arn}'
        cached_creds = await self._redis.hgetall(key)
        if cached_creds:
            return {
                key.decode('utf-8'): value.decode('utf-8')
                for key, value in cached_creds.items()
            }
        try:
            response = await aws.get_credentials(aws_session, role_arn,
                                                 imbi_user.username, id_token)
        except botocore.exceptions.ClientError as error:
            if not is_retry and error.response['Error'][
                    'Code'] == 'ExpiredTokenException':
                id_token = await self.refresh_oauth2_tokens(
                    imbi_user.username, imbi_user.external_id, refresh_token)
                return await self.get_aws_credentials(aws_session, role_arn,
                                                      imbi_user, id_token,
                                                      refresh_token, True)
            elif error.response['Error']['Code'] == 'AccessDenied':
                raise errors.Unauthorized(
                    '%s unauthorized to assume AWS role %s',
                    imbi_user.username,
                    role_arn,
                    reason='Not authorized to access SSM in this '
                    'namespace & environment')
            else:
                raise errors.InternalServerError(
                    'Unexpected AWS credential failure for'
                    '%s assuming AWS role %s: %s',
                    imbi_user.username,
                    role_arn,
                    error.response['Error']['Code'],
                    reason='Unexpected failure fetching AWS credentials')
        creds = {
            'access_key_id': response['Credentials']['AccessKeyId'],
            'secret_access_key': response['Credentials']['SecretAccessKey'],
            'session_token': response['Credentials']['SessionToken']
        }
        await self._redis.hmset_dict(key, creds)
        await self._redis.expireat(
            key, response['Credentials']['Expiration'].timestamp())
        return creds

    async def refresh_oauth2_tokens(self, username, external_id,
                                    refresh_token) -> str:
        google_integration = await oauth2.OAuth2Integration.by_name(
            self.application, self.settings['google']['integration_name'])
        body = urllib.parse.urlencode({
            'grant_type': 'refresh_token',
            'redirect_uri': google_integration.callback_url,
            'refresh_token': refresh_token,
        })
        response = await self.http_fetch(
            str(google_integration.token_endpoint),
            method='POST',
            body=body,
            content_type='application/x-www-form-urlencoded',
            auth_username=google_integration.client_id,
            auth_password=google_integration.client_secret)
        if response.code in (401, 403):
            await self.session.clear()
            self.set_status(403)
            raise web.Finish()
        elif not response.ok:
            self.logger.error(
                'failed to exchange refresh token for new tokens: %s %s',
                response.body['error'], response.body['error_description'])
            raise errors.InternalServerError(
                'failed exchange refresh token for new token: %s',
                response.code,
                title='Google authorization failure',
                instance={
                    'error': response.body['error'],
                    'error_description': response.body['error_description'],
                })
        await google_integration.upsert_user_tokens(
            username,
            external_id,
            response.body['access_token'],
            refresh_token=response.body.get('refresh_token', None),
            id_token=response.body['id_token'])
        return response.body['id_token']


class RecordRequestHandler(SSMRequestHandler, base.PydanticHandlerMixin):
    async def patch(self, *args, **kwargs):
        request_body = self.get_request_body()
        try:
            jsonpatch.JsonPatch(request_body)
        except jsonpatch.InvalidJsonPatch as e:
            raise errors.BadRequest('Invalid JSON patch: %s', e)

        project_info = await self._get_project_info(kwargs['project_id'])
        pydantic_context = {'environments': project_info['environments']}

        patches = self.parse_request_body_as(models.SSMPatchBody,
                                             context=pydantic_context)

        google_tokens = await self._get_google_tokens()
        ssm_path_prefix = path_prefix(
            self.application.settings['project_configuration']
            ['ssm_prefix_template'], project_info['project_slug'],
            project_info['project_type_slug'], project_info['namespace_slug'])
        name = ssm_path_prefix + urllib.parse.unquote(kwargs['name'])
        aws_session = aioboto3.Session()

        for patch in patches:
            patch: models.SSMPatch
            if isinstance(patch, models.RemovePatch):
                await self.delete_param(patch.environment,
                                        project_info['namespace_id'],
                                        aws_session, google_tokens.id_token,
                                        google_tokens.refresh_token, name)
            elif isinstance(patch, models.ReplacePatch):
                if patch.type == 'type':
                    value = await self.get_parameter(
                        patch.environment, project_info['namespace_id'],
                        aws_session, google_tokens.id_token,
                        google_tokens.refresh_token, name)
                    await self.delete_param(patch.environment,
                                            project_info['namespace_id'],
                                            aws_session,
                                            google_tokens.id_token,
                                            google_tokens.refresh_token, name)
                    await self.put_parameter(patch.environment,
                                             project_info['namespace_id'],
                                             aws_session,
                                             google_tokens.id_token,
                                             google_tokens.refresh_token, name,
                                             value, patch.value)
                elif patch.type == 'value':
                    await self.put_parameter(patch.environment,
                                             project_info['namespace_id'],
                                             aws_session,
                                             google_tokens.id_token,
                                             google_tokens.refresh_token,
                                             name,
                                             patch.value,
                                             overwrite=True)
            elif isinstance(patch, models.AddPatch):
                param = patch.value
                await self.put_parameter(patch.environment,
                                         project_info['namespace_id'],
                                         aws_session, google_tokens.id_token,
                                         google_tokens.refresh_token, name,
                                         param.value, param.type)


class CollectionRequestHandler(SSMRequestHandler):
    async def get(self, *args, **kwargs) -> None:
        project_info = await self._get_project_info(kwargs['project_id'])
        if not project_info['environments']:
            self.send_response([])
            return

        google_tokens = await self._get_google_tokens()

        ssm_path_prefix = path_prefix(
            self.application.settings['project_configuration']
            ['ssm_prefix_template'], project_info['project_slug'],
            project_info['project_type_slug'], project_info['namespace_slug'])

        role_arn_exists = False
        params_by_path = collections.defaultdict(dict)
        aws_session = aioboto3.Session()
        for environment in project_info['environments']:
            role_arn = await self.get_role_arn(environment,
                                               project_info['namespace_id'])
            if not role_arn:
                continue
            else:
                role_arn_exists = True

            creds = await self.get_aws_credentials(aws_session, role_arn,
                                                   self.current_user,
                                                   google_tokens.id_token,
                                                   google_tokens.refresh_token)
            params = await aws.get_parameters_by_path(
                aws_session, ssm_path_prefix, creds['access_key_id'],
                creds['secret_access_key'], creds['session_token'])
            for param in params:
                params_by_path[param['Name']][environment] = {
                    'value': param['Value'],
                    'type': param['Type']
                }
        if not role_arn_exists:
            self.logger.info(
                'Fetching SSM params for project %s, no role '
                'ARNs found for namespace %d in any '
                'environment: %s', project_info['project_slug'],
                project_info['namespace_id'], project_info['environments'])
            self.send_response([])
            return

        output = [{
            'name': name,
            'values': {
                environment: {
                    'value': data['value'],
                    'type': data['type']
                }
                for environment, data in d.items()
            },
            'self': self.reverse_url('project-ssm-parameters',
                                     kwargs['project_id']) + '/' +
            urllib.parse.quote(name, safe=[])
        } for name, d in params_by_path.items()]
        self.send_response(output)

    async def post(self, *args, **kwargs) -> None:
        project_info = await self._get_project_info(kwargs['project_id'])
        google_tokens = await self._get_google_tokens()
        aws_session = aioboto3.Session()
        body = self.get_request_body()

        project_config = self.application.settings['project_configuration']
        ssm_path_prefix = path_prefix(project_config['ssm_prefix_template'],
                                      project_info['project_slug'],
                                      project_info['project_type_slug'],
                                      project_info['namespace_slug'])
        name = ssm_path_prefix + body['name']

        for environment, value in body['values'].items():
            await self.put_parameter(environment, project_info['namespace_id'],
                                     aws_session, google_tokens.id_token,
                                     google_tokens.refresh_token, name,
                                     value['value'], value['type'])

    async def delete(self, *args, **kwargs):
        project_info = await self._get_project_info(kwargs['project_id'])
        google_tokens = await self._get_google_tokens()
        aws_session = aioboto3.Session()
        body = self.get_request_body()

        project_config = self.application.settings['project_configuration']
        ssm_path_prefix = path_prefix(project_config['ssm_prefix_template'],
                                      project_info['project_slug'],
                                      project_info['project_type_slug'],
                                      project_info['namespace_slug'])
        name = ssm_path_prefix + body['name']

        for environment in set(body['environments']):
            await self.delete_param(environment, project_info['namespace_id'],
                                    aws_session, google_tokens.id_token,
                                    google_tokens.refresh_token, name)

        self.set_status(204)



"""Nicer application errors.

This module contains specializations of :class:`problemdetails.Problem`
that provide a customizable ``type`` member along with some useful
functionality for specific error cases.

"""
from __future__ import annotations

import typing

import jsonpatch
import jsonpointer
import problemdetails
import pydantic
import tornado.httputil
import tornado.web
import yarl

ERROR_URL = yarl.URL('https://localhost/')


def set_canonical_server(server_name: str) -> None:
    """Call this to override the ``type`` URL in errors."""
    global ERROR_URL
    ERROR_URL = yarl.URL.build(scheme='https', host=server_name)


class ApplicationError(problemdetails.Problem):
    """Generic error with some niceties added.

    The ``log_message`` parameter is required to ensure that we
    have log output when we are returning an error.  There is
    additional code to handle incongruent log format and args
    so that we avoid *most* log format failure exceptions.

    The ``reason`` attribute is always set.  It will use the
    ``title`` parameter or the appropriate HTTP reason.

    The ``type`` property is set to the current ``ERROR_URL``
    with the supplied `fragment` appended.

    The ``detail`` property is set to the formatted log
    message.

    """

    # work around some typing weirdness in problemdetails...
    document: dict[str, typing.Any]
    log_message: str
    reason: str

    def __init__(self, status_code: int, fragment: str, log_message: str,
                 *log_args, **kwargs):
        # NB -- tornado.web.HTTPError DOES NOT set the reason for us :/
        kwargs['reason'] = kwargs.get(
            'reason',
            kwargs.get(
                'title',
                tornado.httputil.responses.get(status_code,
                                               'Unknown Status Code')).title())
        kwargs.setdefault('type', str(ERROR_URL.with_fragment(fragment)))

        # tornado provides partial protection against log message format
        # edge cases but does not insulate against having unnecessary args
        log_args = () if '%' not in log_message else log_args

        super().__init__(status_code, log_message, *log_args, **kwargs)
        self.document.setdefault('title', self.reason)
        self.document.setdefault('detail', self.log_message % log_args)


class BadRequest(ApplicationError):
    def __init__(self, log_message, *log_args, **kwargs):
        super().__init__(400, 'bad-request', log_message, *log_args, **kwargs)


class Unauthorized(ApplicationError):
    def __init__(self, log_message, *log_args, **kwargs):
        super().__init__(401, 'unauthorized', log_message, *log_args, **kwargs)


class Forbidden(ApplicationError):
    def __init__(self, log_message, *log_args, **kwargs):
        super().__init__(403, 'forbidden', log_message, *log_args, **kwargs)


class ClientUnavailableError(Forbidden):
    def __init__(self, client_name: str, log_message: str, *log_args):
        message = log_message % log_args if log_args else log_message
        super().__init__('%s not available: %s', client_name, message)


class ItemNotFound(ApplicationError):
    def __init__(self, log_message=None, *log_args, **kwargs):
        kwargs.setdefault('title', 'Item not found')
        super().__init__(
            404, 'not-found',
            'Item not found' if log_message is None else log_message,
            *log_args, **kwargs)


class MethodNotAllowed(ApplicationError):
    def __init__(self, requested_method: str, **kwargs):
        super().__init__(405, 'method-not-allowed',
                         '%s is not a supported HTTP method',
                         requested_method.upper(), **kwargs)


class UnsupportedMediaType(ApplicationError):
    def __init__(self, media_type: str, **kwargs):
        super().__init__(415, 'unsupported-media-type',
                         '%s is not a supported media type', media_type,
                         **kwargs)


class UnprocessableEntity(ApplicationError):
    def __init__(self, log_message: str, *log_args, **kwargs) -> None:
        fragment = kwargs.pop('fragment', 'unprocessable-entity')
        super().__init__(422, fragment, log_message, *log_args, **kwargs)


class BadJsonPatch(UnprocessableEntity):
    def __init__(self, error: jsonpatch.JsonPatchException
                 | jsonpointer.JsonPointerException, **kwargs) -> None:
        kwargs.setdefault('title', 'Invalid JSON Patch')
        super().__init__('Failed to parse JSON patch: %s', error)


class InternalServerError(ApplicationError):
    def __init__(self, log_message, *log_args, **kwargs):
        super().__init__(500, 'server-error', log_message, *log_args, **kwargs)


class DatabaseError(InternalServerError):
    def __init__(self,
                 log_message=None,
                 *log_args,
                 error: typing.Optional[Exception] = None,
                 **kwargs):
        kwargs.setdefault('title', 'Database Error')
        if log_message is None:
            if error is not None:
                log_message = 'Database failure: %s'
                log_args = (error, )
            else:
                log_message = 'Database failure'
                log_args = ()
        super().__init__(log_message, *log_args, **kwargs)


class IntegrationNotFound(InternalServerError):
    def __init__(self, integration_name: str):
        super().__init__('integration lookup failed for %s',
                         integration_name,
                         title='Integration Missing')


class PydanticValidationError(ApplicationError):
    def __init__(self,
                 error: pydantic.ValidationError,
                 log_message: str,
                 *log_args: str,
                 fragment: str = 'validation-error',
                 status_code: int = 400,
                 **kwargs):
        super().__init__(status_code,
                         fragment,
                         log_message,
                         *log_args,
                         validation_errors=error.errors(include_context=False,
                                                        include_input=False),
                         **kwargs)



import dataclasses
import typing

from imbi import automations, errors, models
from imbi.clients import pagerduty


@dataclasses.dataclass
class PagerDutyContext:
    integration_name: str
    service_info: pagerduty.ServiceInfo


async def create_service(
    context: automations.AutomationContext,
    automation: models.Automation,
    project: models.Project,
) -> None:
    """Create a PagerDuty service for an Imbi project

    This automation creates a new PagerDuty service using the same
    name as `project` and attaching it to the escalation policy
    associated with the namespace. The generated service ID is saved
    as a pagerduty project identifier. A link to the dashboard is
    added to the project if the `automations/pagerduty/project_link_type_id`
    configuration value is set.

    The PagerDuty service information is saved in the context using
    the function object (eg, `create_service`) as the key.

    @see https://developer.pagerduty.com/api-reference/7062f2631b397-create-a-service
    """  # noqa: E501 -- line too long
    if project.namespace.pagerduty_policy is None:
        context.note_progress('not enabled for namespace %s, skipping',
                              project.namespace.name)
        return

    client = await pagerduty.create_client(context.application,
                                           automation.integration_name)
    service = await client.create_service(project)
    context[create_service] = PagerDutyContext(
        integration_name=automation.integration_name, service_info=service)
    context.add_callback(_delete_project)

    # NB -- the integration is a child resource of the service, so it
    # will be deleted when we delete the service, extra cleanup not required
    hook = await client.create_inbound_api_integration(
        'pagerduty-inbound-events', service)
    key = context.application.encrypt_value(hook.integration_key)
    await context.run_query(
        'INSERT INTO v1.project_secrets(project_id, name, value, created_by)'
        '     VALUES (%(project_id)s, %(key_name)s, %(value)s, %(user)s)', {
            'project_id': project.id,
            'key_name': 'pagerduty-integration-key',
            'value': key,
            'user': context.user.username,
        }, 'insert-pagerduty-keys')

    # NB -- the DB fields are cascade deleted when the project is deleted,
    # so extra cleanup is not required
    await context.run_query(
        'INSERT INTO v1.project_identifiers (external_id, integration_name,'
        '                                    project_id, created_by)'
        '     VALUES (%(pagerduty_id)s, %(integration_name)s, %(project_id)s,'
        '             %(user)s)', {
            'pagerduty_id': service.id,
            'integration_name': automation.integration_name,
            'project_id': project.id,
            'user': context.user.username,
        }, 'insert-project-identifiers')
    context.note_progress(
        'registered PagerDuty service %s for Imbi project %s', service.id,
        project.id)

    settings = context.application.settings['automations']['pagerduty']
    if (link_type_id := settings.get('project_link_type_id')) is not None:
        await context.run_query(
            'INSERT INTO v1.project_links (project_id, link_type_id,'
            '                              created_by, url)'
            '     VALUES (%(project_id)s, %(link_type_id)s, %(user)s,'
            '             %(url)s)', {
                'project_id': project.id,
                'link_type_id': link_type_id,
                'user': context.user.username,
                'url': str(service.html_url),
            }, 'insert-project-links')
        context.note_progress('created PagerDuty link %s for Imbi project %s',
                              service.html_url, project.id)


async def _delete_project(context: automations.AutomationContext,
                          _error: BaseException) -> None:
    try:
        pd_info = typing.cast(PagerDutyContext, context[create_service])
    except KeyError:
        return

    try:
        client = await pagerduty.create_client(context.application,
                                               pd_info.integration_name)
    except errors.ClientUnavailableError:
        pass
    else:
        context.note_progress('removing PagerDuty service %s due to error',
                              pd_info.service_info.id)
        await client.remove_service(pd_info.service_info.id)


async def associate_service_dependency(
    context: automations.AutomationContext,
    automation: models.Automation,
    dependency: models.ProjectDependency,
) -> None:
    if not (dependent_service_id := await _get_service_id(
            context, dependency.project_id)):
        context.note_progress(
            'Pagerduty service not set up for '
            'project %d, skipping automation', dependency.project_id)
        return

    if not (dependency_service_id := await _get_service_id(
            context, dependency.dependency_id)):
        context.note_progress(
            'Pagerduty service not set up for '
            'project %d, skipping automation', dependency.dependency_id)
        return

    client = await pagerduty.create_client(context.application,
                                           automation.integration_name)

    service_dependencies = await client.get_service_dependencies(
        dependent_service_id)
    for d in service_dependencies:
        if d.supporting_service.id == dependency_service_id:
            context.note_progress(
                'Pagerduty service dependency already exists '
                'between %s and %s, skipping automation',
                dependency.project_id, dependency.dependency_id)
            return

    await client.associate_service_dependency(
        supporting_service_id=dependency_service_id,
        dependent_service_id=dependent_service_id)


async def disassociate_service_dependency(
    context: automations.AutomationContext,
    automation: models.Automation,
    dependency: models.ProjectDependency,
) -> None:
    if not (dependent_service_id := await _get_service_id(
            context, dependency.project_id)):
        context.note_progress(
            'Pagerduty service not set up for '
            'project %d, skipping automation', dependency.project_id)
        return

    if not (dependency_service_id := await _get_service_id(
            context, dependency.dependency_id)):
        context.note_progress(
            'Pagerduty service not set up for '
            'project %d, skipping automation', dependency.dependency_id)
        return

    client = await pagerduty.create_client(context.application,
                                           automation.integration_name)

    service_dependencies = await client.get_service_dependencies(
        dependent_service_id)
    for d in service_dependencies:
        if d.supporting_service.id == dependency_service_id:
            await client.disassociate_service_dependency(
                supporting_service_id=dependency_service_id,
                dependent_service_id=dependent_service_id)
            return
    context.note_progress(
        'Pagerduty service dependency does not exist '
        'between %s and %s, skipping automation', dependency.project_id,
        dependency.dependency_id)


async def _get_service_id(context: automations.AutomationContext,
                          project_id: int) -> typing.Optional[str]:
    result = result = await context.run_query(
        'SELECT external_id'
        '  FROM v1.project_identifiers'
        ' WHERE project_id = %(project_id)s'
        '   AND integration_name = %(integration_name)s', {
            'project_id': project_id,
            'integration_name': context.current_integration,
        },
        metric_name='get-pagerduty-service-id')
    if not result.row_count:
        return None
    return result.row['external_id']



import decimal
import urllib.parse

import pydantic
from sprockets.mixins.mediatype import handlers, transcoders


def parse_form_body(data) -> dict:
    def translate_value(v):
        if not v:
            return None

        m = {'null': None, 'true': True, 'false': False}
        try:
            value = m[v]
        except KeyError:
            try:
                value = float(v)
            except ValueError:
                value = v
        return value

    if hasattr(data, 'decode'):
        data = data.decode('utf-8')
    form_body = {}
    parsed = urllib.parse.parse_qsl(data, keep_blank_values=True)
    for name, value in parsed:
        value = translate_value(value)
        if name in form_body:
            if not isinstance(form_body[name], list):
                form_body[name] = [form_body[name]]
            form_body[name].append(value)
        else:
            form_body[name] = value
    return form_body


class DecimalMixin:
    def dump_object(self, obj):
        if isinstance(obj, decimal.Decimal):
            return float(obj)
        return super().dump_object(obj)


class PydanticMixin:
    def dump_object(self, obj):
        if isinstance(obj, pydantic.BaseModel):
            return obj.model_dump(mode='json')
        return super().dump_object(obj)


class JSONTranscoder(DecimalMixin, PydanticMixin, transcoders.JSONTranscoder):
    """Handle Decimal Values"""


class MsgPackTranscoder(DecimalMixin, PydanticMixin,
                        transcoders.MsgPackTranscoder):
    """Handle Decimal Values"""


class HTMLTranscoder(transcoders.JSONTranscoder):
    def __init__(self, content_type='text/html', default_encoding='utf-8'):
        super().__init__(content_type, default_encoding)
        self.dump_options = {
            'default': self.dump_object,
            'separators': (',', ':')
        }

    def dumps(self, value):
        """Just pass through the value if it's a string, otherwise dump it as
        JSON.

        :rtype: str

        """
        if not isinstance(value, (bytes, str)):
            value = '<html><body><pre>{}</pre></body></html>'.format(
                super().dumps(value))
        return value

    @staticmethod
    def loads(value):
        """Just pass through the value."""
        return value


class FormTranscoder(handlers.TextContentHandler):
    def __init__(self):
        super().__init__('application/x-www-form-urlencoded',
                         dumps=self.dumps,
                         loads=self.loads,
                         default_encoding='utf-8')

    def dumps(self, inst_data) -> str:
        return urllib.parse.urlencode(inst_data)

    def loads(self, data):
        return parse_form_body(data)



import re
import uuid

import ulid
from tornado import web

from . import base


class RequestHandler(base.ValidatingRequestHandler):

    NAME = 'authentication-tokens'

    CREATE_SQL = re.sub(
        r'\s+', ' ', """
        INSERT INTO v1.authentication_tokens (token, "name", username)
             VALUES (%(token)s, %(name)s, %(username)s)
          RETURNING username, token, "name",
                    created_at, expires_at, last_used_at
        """)

    DELETE_SQL = re.sub(
        r'\s+', ' ', """
        DELETE FROM v1.authentication_tokens
              WHERE username = %(username)s
                AND token = %(token)s""")

    GET_SQL = re.sub(
        r'\s+', ' ', """
        SELECT token, "name", username, created_at, expires_at, last_used_at
          FROM v1.authentication_tokens
         WHERE username = %(username)s
         ORDER BY created_at""")

    async def delete(self, token):
        result = await self.postgres_execute(self.DELETE_SQL, {
            'token': token,
            'username': self.current_user.username
        }, 'delete-authentication-token')
        if not result.row_count:
            raise web.HTTPError(404, reason='Item not found')
        self.set_status(204, reason='Item Deleted')

    async def get(self):
        result = await self.postgres_execute(
            self.GET_SQL, {'username': self.current_user.username},
            'get-authentication-tokens')
        self.send_response(result.rows)

    async def post(self):
        values = self.get_request_body()
        values.update({
            'token': uuid.UUID(ulid.ULID().hex),
            'username': self.current_user.username
        })
        result = await self.postgres_execute(self.CREATE_SQL, values,
                                             'create-authentication-token')
        self.send_response(result.row)



import typing
import re

from imbi import automations, models, errors
from imbi.endpoints import base, projects


class _DependencyRequestMixin:

    ID_KEY = ['project_id', 'dependency_id']
    ITEM_NAME = 'project-dependency'
    FIELDS = ['project_id', 'dependency_id']
    TTL = 300

    GET_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT project_id, created_at, created_by, dependency_id
          FROM v1.project_dependencies
         WHERE project_id=%(project_id)s
           AND dependency_id=%(dependency_id)s""")

    POST_SQL = re.sub(
        r'\s+', ' ', """\
        INSERT INTO v1.project_dependencies
                    (project_id, dependency_id, created_by)
             VALUES (%(project_id)s, %(dependency_id)s, %(username)s)
          RETURNING project_id, dependency_id, created_at, created_by""")

    async def _run_automations(
            self, dependency: models.ProjectDependency,
            selected_automations: typing.Sequence[models.Automation]) -> None:
        """Run a list of automations for the dependency

        If an automation fails, then an InternalServerError is raised.
        """
        if not selected_automations:
            return

        ordered_automations = automations.sort_automations(
            selected_automations)

        self.logger.info(
            'running automations for dependency '
            '(project: %s, dependency: %s): %r', dependency.project_id,
            dependency.dependency_id, [a.slug for a in ordered_automations])
        try:
            await automations.run_automations(ordered_automations,
                                              dependency,
                                              application=self.application,
                                              user=self._current_user,
                                              query_executor=self)
        except errors.ApplicationError:  # this is meant for the end user
            raise
        except automations.AutomationFailedError as error:
            raise errors.InternalServerError(str(error)) from None


class CollectionRequestHandler(projects.ProjectAttributeCollectionMixin,
                               _DependencyRequestMixin,
                               base.CollectionRequestHandler):

    NAME = 'project-dependencies'

    COLLECTION_SQL = re.sub(
        r'\s+', ' ', """\
          SELECT project_id, created_by, dependency_id
            FROM v1.project_dependencies
           WHERE project_id=%(project_id)s
        ORDER BY dependency_id""")

    COLLECTION_WITH_DEPENDENCY_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT d.project_id, d.created_at, d.created_by, d.dependency_id,
               p.name AS dependency_name,
               p.namespace_id AS dependency_namespace_id,
               p.project_type_id AS dependency_project_type_id
          FROM v1.project_dependencies AS d
          JOIN v1.projects AS p
            ON d.dependency_id=p.id
         WHERE project_id=%(project_id)s""")

    async def get(self, *args, **kwargs):
        if 'dependency' in self.get_query_arguments('include'):
            result = await self.postgres_execute(
                self.COLLECTION_WITH_DEPENDENCY_SQL,
                kwargs,
                metric_name='get-{}'.format(self.NAME))
            response = [{
                'project_id': row['project_id'],
                'dependency_id': row['dependency_id'],
                'created_at': row['created_at'],
                'created_by': row['created_by'],
                'dependency': {
                    'id': row['dependency_id'],
                    'name': row['dependency_name'],
                    'namespace_id': row['dependency_namespace_id'],
                    'project_type_id': row['dependency_project_type_id'],
                }
            } for row in result.rows]
            self.send_response(response)
        else:
            await super().get(*args, **kwargs)

    async def post(self, *args, **kwargs):
        values = self.get_request_body()
        values['project_id'] = kwargs['project_id']
        values['username'] = self._current_user.username
        dependent_project = await models.project(values['project_id'],
                                                 self.application)
        selected_automations = await automations.retrieve_automations(
            self.application, values.get('automations', []),
            dependent_project.project_type.id)

        result = await self.postgres_execute(self.POST_SQL, values,
                                             f'post-{self.NAME}')
        if not result.row_count:
            raise errors.DatabaseError('Failed to create project dependency',
                                       title='Failed to create record')

        dependency = models.ProjectDependency(
            project_id=result.row['project_id'],
            dependency_id=result.row['dependency_id'],
            created_at=result.row['created_at'],
            created_by=result.row['created_by'])

        try:
            await self._run_automations(dependency, selected_automations)
        except Exception as error:
            self.logger.exception('_run_automations failure: %s', error)
            self.logger.error(
                'removing dependency (project: %s, '
                'dependency: %s) due to error', dependency.project_id,
                dependency.dependency_id)
            await self.postgres_execute(
                'DELETE FROM v1.project_dependencies'
                '     WHERE project_id = %(project_id)s'
                '       AND dependency_id = %(dependency_id)s', {
                    'project_id': dependency.project_id,
                    'dependency_id': dependency.dependency_id
                })
            raise errors.InternalServerError('Failed to run automations: %s',
                                             error) from None

        await self.index_document(dependent_project.id)
        self.send_response(vars(dependency))


class RecordRequestHandler(projects.ProjectAttributeCRUDMixin,
                           _DependencyRequestMixin, base.CRUDRequestHandler):

    NAME = 'project-dependency'

    DELETE_SQL = re.sub(
        r'\s+', ' ', """\
        DELETE FROM v1.project_dependencies
         WHERE project_id=%(project_id)s
           AND dependency_id=%(dependency_id)s""")

    async def delete(self, *args, **kwargs):
        body = self.get_request_body() if self.request.body else {}
        dependency = await models.project_dependency(kwargs['project_id'],
                                                     kwargs['dependency_id'],
                                                     self.application)
        if dependency is None:
            raise errors.ItemNotFound(instance=self.request.uri)
        dependent_project = await models.project(kwargs['project_id'],
                                                 self.application)
        selected_automations = await automations.retrieve_automations(
            self.application, body.get('automations', []),
            dependent_project.project_type.id)

        result = await self.postgres_execute(self.DELETE_SQL, kwargs,
                                             f'delete-{self.NAME}')
        if not result.row_count:
            raise errors.ItemNotFound(instance=self.request.uri)

        try:
            await self._run_automations(dependency, selected_automations)
        except Exception as error:
            self.logger.exception('_run_automations failure: %s', error)
            self.logger.error(
                're-adding dependency (project: %s, '
                'dependency: %s) due to error', dependency.project_id,
                dependency.dependency_id)
            await self.postgres_execute(
                self.POST_SQL, {
                    'project_id': dependency.project_id,
                    'dependency_id': dependency.dependency_id,
                    'created_by': dependency.created_by
                })
            raise errors.InternalServerError('Failed to run automations: %s',
                                             error) from None

        await self.index_document(kwargs['project_id'])
        self.set_status(204, reason='Item Deleted')



import re

from imbi.endpoints import base


class CollectionRequestHandler(base.CollectionRequestHandler):
    NAME = 'integrations'
    ITEM_NAME = 'integration'
    ID_KEY = 'name'
    FIELDS = ['api_endpoint', 'api_secret']

    COLLECTION_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT name, api_endpoint,
               CASE api_secret
                    WHEN NULL THEN NULL
                    ELSE LEFT(api_secret, 3) || '...' || RIGHT(api_secret, 3)
               END AS api_secret
          FROM v1.integrations
         ORDER BY name
        """)

    GET_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT name, api_endpoint,
               CASE api_secret
                    WHEN NULL THEN NULL
                    ELSE LEFT(api_secret, 3) || '...' || RIGHT(api_secret, 3)
               END AS api_secret
          FROM v1.integrations
         WHERE name = %(name)s
        """)

    POST_SQL = re.sub(
        r'\s+', ' ', """\
        INSERT INTO v1.integrations
                    (name, api_endpoint, api_secret, created_at, created_by)
             VALUES (%(name)s, %(api_endpoint)s, %(api_secret)s,
                     CURRENT_TIMESTAMP, %(username)s)
          RETURNING name, api_endpoint
        """)

    @base.require_permission('admin')
    async def post(self, *args, **kwargs) -> None:
        await super().post(*args, **kwargs)


class RecordRequestHandler(base.CRUDRequestHandler):
    NAME = 'integration'
    ID_KEY = 'name'

    GET_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT name, api_endpoint, created_at, created_by,
               last_modified_at, last_modified_by,
               CASE api_secret
                    WHEN NULL THEN NULL
                    ELSE LEFT(api_secret, 3) || '...' || RIGHT(api_secret, 3)
               END AS api_secret
          FROM v1.integrations
         WHERE name = %(name)s
        """)

    PATCH_SQL = re.sub(
        r'\s+', ' ', """\
        UPDATE v1.integrations
           SET api_endpoint = %(api_endpoint)s,
               api_secret = %(api_secret)s,
               last_modified_by = %(username)s,
               last_modified_at = CURRENT_TIMESTAMP
         WHERE name = %(name)s
           AND name = %(current_name)s
        RETURNING api_endpoint
        """)

    DELETE_SQL = re.sub(
        r'\s+', ' ', """\
           DELETE FROM v1.integrations
                 WHERE name = %(name)s
           """)

    @base.require_permission('admin')
    async def delete(self, *args, **kwargs) -> None:
        await super().delete(*args, **kwargs)

    @base.require_permission('admin')
    async def patch(self, *args, **kwargs) -> None:
        await super().patch(*args, **kwargs)



from __future__ import annotations

import dataclasses
import logging
import typing
import urllib.parse

import sprockets.mixins.http
import yarl

from imbi import errors, models, version
if typing.TYPE_CHECKING:
    from imbi import app


@dataclasses.dataclass
class ProjectInfo:
    key: str
    dashboard_url: yarl.URL


def generate_key(project: models.Project) -> str:
    """Generate a SonarQube project key for `project`."""
    return ':'.join([project.namespace.slug.lower(), project.slug.lower()])


async def create_client(application: app.Application,
                        integration_name: str) -> '_SonarQubeClient':
    """Create a SonarQube API client

    The settings needed to create a client are in the application
    settings and the database currently. This function finds them
    and creates a client.  If the configuration is invalid or the
    integration is disabled, a ``ClientUnavailableError`` is raised
    with an appropriate message.
    """
    logger = logging.getLogger(__package__).getChild('create_client')
    settings = application.settings['automations']['sonarqube']
    if not settings['enabled']:
        raise errors.ClientUnavailableError(integration_name, 'disabled')

    sonarqube_info = await models.integration(integration_name, application)
    if not sonarqube_info:
        logger.warning('%r integration is enabled but not configured',
                       integration_name)
        raise errors.ClientUnavailableError(integration_name, 'not configured')
    if not sonarqube_info.api_secret:
        logger.warning('API secret is missing for %r', integration_name)
        raise errors.ClientUnavailableError(integration_name, 'misconfigured')

    return _SonarQubeClient(yarl.URL(str(sonarqube_info.api_endpoint)),
                            sonarqube_info.api_secret)


class _SonarQubeClient(sprockets.mixins.http.HTTPClientMixin):
    gitlab_alm_key: typing.Union[bool, None, str] = None
    """Cached ALM key for gitlab

    If we haven't checked the ALM configuration in the sonar server,
    then this is `None`. If we have retrieved the ALM configuration
    that this is either `False` if it is disabled or the integration
    key (as a str) if it is enabled.
    """
    def __init__(self, api_endpoint: yarl.URL, api_secret: str, *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.logger = logging.getLogger(__package__).getChild(
            'SonarQubeClient')
        self.api_url = api_endpoint
        self.api_secret = api_secret
        self.url = api_endpoint.with_path('/')

    async def api(self,
                  url: typing.Union[yarl.URL, str],
                  *,
                  method: str = 'GET',
                  **kwargs) -> sprockets.mixins.http.HTTPResponse:
        """Make an authenticated API call."""
        if not isinstance(url, yarl.URL):
            url = yarl.URL(url)
        if not url.is_absolute():
            new_url = self.api_url / url.path.lstrip('/')
            url = new_url.with_query(url.query)

        request_headers = kwargs.setdefault('request_headers', {})
        request_headers['Accept'] = 'application/json'
        if kwargs.get('body', None) is not None:
            kwargs['content_type'] = 'application/x-www-form-urlencoded'
            request_headers['Content-Type'] = kwargs['content_type']
            body = urllib.parse.urlencode(kwargs.pop('body'))
            kwargs['body'] = body.encode()

        kwargs.update({
            'auth_username': self.api_secret,
            'auth_password': '',
            'user_agent': f'imbi/{version} (SonarQubeClient)',
        })
        response = await super().http_fetch(str(url), method=method, **kwargs)
        if not response.ok:
            self.logger.warning('%s %s failed: %s', method, url, response.code)
            if response.body:
                self.logger.warning('response body: %r', response.body)
        return response

    async def create_project(self, project: models.Project) -> ProjectInfo:
        """Create a SonarQube project for `project`.

        :returns: a :class:`tuple` containing the assigned project key
            and a link to the dashboard

        """
        self.logger.info('creating SonarQube project for %s', project.slug)
        response = await self.api('/projects/create',
                                  method='POST',
                                  body={
                                      'name': project.name,
                                      'project': generate_key(project),
                                  })
        if not response.ok:
            raise errors.InternalServerError('failed to create project %s: %s',
                                             project.name,
                                             response.code,
                                             title='SonarQube API Failure',
                                             sonar_response=response.body)
        project_key = response.body['project']['key']
        dashboard_url = self.url.with_path('/dashboard').with_query({
            'id': project_key,
        })
        return ProjectInfo(key=project_key, dashboard_url=dashboard_url)

    async def remove_project(self, project: ProjectInfo) -> None:
        """Remove an existing project"""
        await self.api('/projects/delete', body={'project': project.key})

    async def enable_pr_decoration(self, project: ProjectInfo,
                                   gitlab_project_id: int):
        """Enable GitLab MR decoration if it is available

        The MR decoration settings contain the gitlab project ID.
        This is used to post the analysis results to an MR when
        an analysis runs. The GitLab application lifecycle
        management (ALM) plugin in SonarQube may be unconfigured
        or unavailable. The internal `_is_gitlab_alm_available`
        method checks if it is configured at the system level.
        We need to set the ALM binding for the project to the
        GitLab project ID if SonarQube is configured to update
        GitLab.

        """
        alm_enabled = await self._is_gitlab_alm_available()
        if not alm_enabled:
            return

        if not gitlab_project_id:
            self.logger.error('refusing to set PR repository to %r',
                              gitlab_project_id)
            return

        self.logger.debug('checking for PR decoration on %s', project.key)
        response = await self.api(
            yarl.URL('/alm_settings/get_binding').with_query({
                'project': project.key,
            }))
        if response.code == 404:  # not configured or doesn't exist
            self.logger.debug(
                'setting PR decoration for %s: almSetting=%r repository=%r',
                project.key, self.gitlab_alm_key, gitlab_project_id)
            response = await self.api(
                yarl.URL('/alm_settings/set_gitlab_binding'),
                method='POST',
                body={
                    'almSetting': self.gitlab_alm_key,
                    'project': project.key,
                    'repository': gitlab_project_id,
                })
            if not response.ok:
                self.logger.error('PR decoration failure %s: %s', project.key,
                                  response.code)
                raise errors.InternalServerError(
                    'failed to enable PR decoration for %s: %s',
                    project.key,
                    response.code,
                    title='SonarQube Client Failure')
        elif not response.ok:
            raise errors.InternalServerError(
                'failed to retrieve ALB bindings for %s: %s',
                project.key,
                response.code,
                title='SonarQube Client Failure')

    async def _is_gitlab_alm_available(self) -> bool:
        """Check if the SonarQube server has the GitLab ALM configured.

        The result of this method is "remembered" by setting the
        `gitlab_alm_key` class attribute to something other than
        :data:`None`.  It will be set to :data:`False` when we
        determine that the GitLab ALM is not configured or to the
        configured "key" value.

        """
        if self.gitlab_alm_key is None:
            response = await self.api('/alm_settings/list_definitions')
            if response.ok:
                self.__class__.gitlab_alm_key = False
                for connection in response.body.get('gitlab', []):
                    if connection['key'] == 'gitlab':
                        self.logger.info('found GitLab ALM %s',
                                         connection['key'])
                        self.__class__.gitlab_alm_key = connection['key']
                        break
                else:
                    self.logger.warning(
                        'GitLab ALM not found in %r, disabling MR decoration',
                        response.body)
            else:
                self.logger.warning(
                    'failed to list alm_settings for sonar: %s', response.code)
                return False  # don't cache failures for now

        return bool(self.gitlab_alm_key)



import dataclasses
import typing

from imbi import automations, errors, models
from imbi.clients import sentry


@dataclasses.dataclass
class SentryContext:
    integration_name: str
    project_info: sentry.ProjectInfo


async def create_project(
    context: automations.AutomationContext,
    automation: models.Automation,
    project: models.Project,
) -> None:
    """Create a Sentry project for an Imbi project

    This automation creates a new Sentry project, adds the sentry
    slug as an object identifier, and sets the Sentry link if one
    is configured in settings. The project is initially added to
    the Sentry team associated with the Project namespace.

    The Sentry client retrieves the secret and public client DSNs
    after it creates the project. The public DSN is stored as a
    client secret.

    The sentry project information (as returned by the Sentry
    client) is stored in the context using the function object
    (eg, `create_project`) as the key.

    @see https://docs.sentry.io/api/projects/create-a-new-project/
    @see https://docs.sentry.io/api/projects/list-a-projects-client-keys/

    """
    client = await sentry.create_client(context.application,
                                        automation.integration_name)
    sentry_project = await client.create_project(
        project.namespace.sentry_team_slug or project.namespace.slug,
        project.name)
    context[create_project] = SentryContext(
        integration_name=automation.integration_name,
        project_info=sentry_project)
    context.add_callback(_delete_project)
    context.note_progress('created Sentry project %s for Imbi project %s',
                          sentry_project.slug, project.id)

    # NB -- project is deleted on failure and the deletion is
    # cascaded to the v1.project_secrets, v1.project_identifiers,
    # and v1.project_links tables, so we do not need an explicit cleanup

    key = context.application.encrypt_value(str(sentry_project.keys['public']))
    await context.run_query(
        'INSERT INTO v1.project_secrets(project_id, name, value, created_by)'
        '     VALUES (%(project_id)s, %(key_name)s, %(value)s, %(user)s)', {
            'project_id': project.id,
            'key_name': 'sentry-public-dsn',
            'value': key,
            'user': context.user.username,
        }, 'insert-sentry-keys')
    context.note_progress('stored Public Sentry DSN for Imbi project %s',
                          project.id)

    await context.run_query(
        'INSERT INTO v1.project_identifiers (external_id, integration_name,'
        '                                    project_id, created_by)'
        '     VALUES (%(sentry_id)s, %(integration_name)s, %(project_id)s,'
        '             %(user)s)', {
            'sentry_id': sentry_project.slug,
            'integration_name': automation.integration_name,
            'project_id': project.id,
            'user': context.user.username
        }, 'insert-project-identifiers')
    context.note_progress(
        'registered Sentry identifier %s for Imbi project %s',
        sentry_project.slug, project.id)

    settings = context.application.settings['automations']['sentry']
    if (link_type_id := settings.get('project_link_type_id')) is not None:
        await context.run_query(
            'INSERT INTO v1.project_links (project_id, link_type_id,'
            '                              created_by, url)'
            '     VALUES (%(project_id)s, %(link_type_id)s, %(user)s,'
            '             %(dashboard_url)s)', {
                'project_id': project.id,
                'link_type_id': link_type_id,
                'user': context.user.username,
                'dashboard_url': str(sentry_project.link)
            }, 'insert-project-links')
        context.note_progress('created Sentry link %s for Imbi project %s',
                              sentry_project.link, project.id)


async def _delete_project(context: automations.AutomationContext,
                          _error: BaseException) -> None:
    try:
        sentry_info = typing.cast(SentryContext, context[create_project])
    except KeyError:
        return

    try:
        client = await sentry.create_client(context.application,
                                            sentry_info.integration_name)
    except errors.ClientUnavailableError:
        pass
    else:
        context.note_progress('removing sentry project %s due to error',
                              sentry_info.project_info.slug)
        await client.remove_project(sentry_info.project_info.slug)



import hmac
import random
import string
import unittest

import imbi.keychain


class KeychainCreationTests(unittest.TestCase):
    def test_that_32_byte_key_is_required(self):
        with self.assertRaises(ValueError):
            imbi.keychain.Keychain(b'not 32 bytes')


class PasswordHashingTests(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.key = b'some thirty-two character secret'

    def test_that_hash_uses_secret_as_hmac_key(self):
        encrypter = hmac.new(self.key, digestmod='SHA512')
        encrypter.update(b'my secret')
        keychain = imbi.keychain.Keychain(self.key)
        self.assertEqual(encrypter.digest(), keychain.hash('my secret'))


class EncryptionTests(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.key = b'some thirty-two character secret'

    def test_symmetric_encryption(self):
        # generate some random "words"
        plaintext = ' '.join(''.join(
            random.choice(string.ascii_letters)
            for _ in range(random.randint(5, 10)))
                             for _ in range(random.randint(10, 15))).encode()

        # round trip them through the keychain
        keychain = imbi.keychain.Keychain(self.key)
        ciphertext = keychain.encrypt(plaintext)
        self.assertEqual(plaintext, keychain.decrypt(ciphertext))

    def test_decryption_with_tampered_ciphertext(self):
        keychain = imbi.keychain.Keychain(self.key)
        ciphertext = keychain.encrypt(b'some plain text')
        ciphertext = ciphertext[5:] + ciphertext[:5]
        with self.assertRaises(imbi.keychain.DecryptionFailure):
            keychain.decrypt(ciphertext)

    def test_that_keychains_with_same_key_can_roundtrip(self):
        plaintext = b'my dirty little secret'
        keychain1 = imbi.keychain.Keychain(self.key)
        ciphertext = keychain1.encrypt(plaintext)

        keychain2 = imbi.keychain.Keychain(self.key)
        self.assertEqual(plaintext, keychain2.decrypt(ciphertext))



from imbi.endpoints import base
from imbi.endpoints.components import models


class RequestHandler(base.AuthenticatedRequestHandler):
    NAME = 'reports-outdated-projects'

    async def get(self) -> None:
        result = await self.postgres_execute(
            'SELECT c.project_id, v.status, p.name AS project_name,'
            '       n.name AS project_namespace,'
            '       COUNT(c.version_id) AS num_components'
            '  FROM v1.project_components AS c'
            '  JOIN v1.component_versions AS v'
            '    ON v.package_url = c.package_url AND v.id = c.version_id'
            '  JOIN v1.projects AS p ON p.id = c.project_id'
            '  JOIN v1.namespaces AS n ON n.id = p.namespace_id'
            ' GROUP BY c.project_id, v.status, p.name, n.name'
            ' ORDER BY c.project_id',
            metric_name=self.NAME)

        # using a helper here makes it easy to pick up the
        # values from the last set of rows
        def add_current_to_report() -> None:
            outdated = sum(
                current.get(k, 0) for k in models.OUTDATED_COMPONENT_STATUSES)
            if outdated:
                current['component_score'] = models.ProjectStatus.calculate(
                    current)
                # pivot from 'Up-to-date' to 'up_to_date'
                for e in models.ProjectComponentStatus:
                    current[e.name.lower()] = current.pop(e.value)
                report_rows.append(current.copy())

        report_rows = []
        current = {}
        empty = {e.value: 0 for e in models.ProjectComponentStatus}

        for row in result.rows:
            if row['project_id'] != current.get('project_id'):
                add_current_to_report()
                current = empty | {
                    'project_id': row['project_id'],
                    'project_namespace': row['project_namespace'],
                    'project_name': row['project_name']
                }
            current[row['status']] = row['num_components']
        add_current_to_report()  # capture the final in-progress row

        self.send_response(report_rows)



"""
System Reports

"""
from tornado import web

from . import authentication, index, settings, user

IndexRequestHandler = index.IndexRequestHandler

URLS = [
    web.url(r'^/ui/', index.IndexRequestHandler),
    web.url(r'/ui/connections/(?P<integration_name>.*)',
            authentication.ConnectionRequestHandler),
    web.url(r'^/ui/login$', authentication.LoginRequestHandler),
    web.url(r'^/ui/login/google$', authentication.GoogleLoginRequestHandler),
    web.url(r'^/ui/logout$', authentication.LogoutRequestHandler),
    web.url(r'^/ui/settings$', settings.RequestHandler),
    web.url(r'^/ui/user$', user.UserRequestHandler),
    web.url(r'^/ui/available-automations$', user.AvailableAutomationsHandler),
    web.url(r'^/ui/.*$', index.IndexRequestHandler)
]



from tornado import web

from . import handlers

URLS = [
    web.url(r'^/projects/(?P<project_id>\d+)/configuration/ssm$',
            handlers.CollectionRequestHandler,
            name='project-ssm-parameters'),
    web.url(
        r'^/projects/(?P<project_id>\d+)/configuration/ssm/'
        r'(?P<name>[\w_\-%\+]+)$',
        handlers.RecordRequestHandler,
        name='project-ssm-parameter'),
]



from __future__ import annotations

import datetime
import enum
import inspect
import logging
import typing

import pydantic

from . import do_nothing
from imbi import errors, slugify

if typing.TYPE_CHECKING:
    from imbi import app

__all__ = [
    'Automation',
    'AutomationCategory',
    'automation',
    'Integration',
    'integration',
]

LOGGER = logging.getLogger(__name__)


def verify_legal_callable(v):
    if not inspect.iscoroutinefunction(v):
        name = getattr(v, '__name__', repr(v))
        raise ValueError(f'{name} is not a coroutine function')

    mod_name = inspect.getmodule(v).__name__
    allow_list = {'imbi.automations'}
    if (mod_name in allow_list
            or any(mod_name.startswith(a + '.') for a in allow_list)):
        return v
    raise ValueError(f'{mod_name!r} is not an allowed implementation module')


CallableType = typing.Annotated[pydantic.ImportString,
                                pydantic.AfterValidator(verify_legal_callable),
                                pydantic.Field(default=do_nothing)]
PathIdType: typing.TypeAlias = typing.Union[int, slugify.Slug]


class AutomationCategory(enum.Enum):
    CREATE_PROJECT = 'create-project'
    CREATE_DEPENDENCY = 'create-project-dependency'
    REMOVE_DEPENDENCY = 'remove-project-dependency'


class Automation(pydantic.BaseModel):
    id: int
    name: str
    slug: slugify.Slug
    integration_name: str
    callable: CallableType
    categories: list[AutomationCategory] = pydantic.Field(min_length=1)
    applies_to: list[slugify.Slug] = pydantic.Field(default_factory=list,
                                                    min_length=1)
    applies_to_ids: list[int] = pydantic.Field(default_factory=list,
                                               min_length=1,
                                               exclude=True)
    depends_on: list[slugify.Slug] = pydantic.Field(default_factory=list)
    depends_on_ids: list[int] = pydantic.Field(default_factory=list,
                                               exclude=True)
    created_by: str
    created_at: datetime.datetime
    last_modified_by: typing.Union[str, None] = None
    last_modified_at: typing.Union[datetime.datetime, None] = None

    @pydantic.field_validator('categories', mode='before')
    @classmethod
    def handle_postgres_array_syntax(cls, value) -> list[AutomationCategory]:
        if isinstance(value, str):
            return [AutomationCategory(v) for v in value[1:-1].split(',')]
        return value

    @pydantic.field_validator('applies_to',
                              'applies_to_ids',
                              'depends_on',
                              'depends_on_ids',
                              mode='before')
    @classmethod
    def handle_postgres_null_arrays(cls,
                                    value) -> list[slugify.Slug] | list[int]:
        """Adjust for postgres array values...

        The `array_agg()` function has some interesting behavior
        when mixed with `LEFT JOIN` like returning an array of
        `NULL` values when there is not a match. Other references
        can cause the array itself to be `NULL`. In our specific
        use case, we always want an array, and it should never
        contain `None`.

        """
        return [v for v in value if v is not None and v is not None]


class Integration(pydantic.BaseModel):
    name: str
    api_endpoint: pydantic.HttpUrl
    api_secret: typing.Union[str, None] = None
    created_at: datetime.datetime
    created_by: str
    last_modified_at: typing.Union[datetime.datetime, None] = None
    last_modified_by: typing.Union[str, None] = None


async def automation(automation_slug: str,
                     application: 'app.Application') -> Automation | None:
    def on_postgres_error(_metric_name: str, exc: Exception) -> None:
        LOGGER.error('Failed to execute query for automation %s: %s',
                     automation_slug, exc)
        raise errors.DatabaseError('Error loading Automation', error=exc)

    async with application.postgres_connector(
            on_error=on_postgres_error) as conn:
        result = await conn.execute(
            'SELECT a.id, a.name, a.slug, a.integration_name, a.callable,'
            '       a.categories, a.created_at, a.created_by,'
            '       a.last_modified_at, a.last_modified_by,'
            '       array_agg(DISTINCT pt.slug) AS applies_to,'
            '       array_agg(DISTINCT d.slug) AS depends_on,'
            '       array_agg(DISTINCT pt.id) AS applies_to_ids,'
            '       array_agg(DISTINCT d.id) AS depends_on_ids'
            '  FROM      v1.automations AS a'
            '  LEFT JOIN v1.available_automations AS aa'
            '         ON aa.automation_id = a.id'
            '  LEFT JOIN v1.project_types AS pt ON pt.id = aa.project_type_id'
            '  LEFT JOIN v1.automations_graph AS g ON g.automation_id = a.id'
            '  LEFT JOIN v1.automations AS d ON d.id = g.dependency_id'
            ' WHERE a.slug = %(automation_slug)s'
            ' GROUP BY a.id, a.name, a.callable, a.categories, a.slug,'
            '          a.integration_name, a.created_at, a.created_by,'
            '          a.last_modified_at, a.last_modified_by',
            {'automation_slug': automation_slug})
        if result.row_count:
            return Automation.model_validate(result.row)


async def integration(integration_name: str,
                      application: 'app.Application') -> Integration | None:
    def on_postgres_error(_metric_name: str, exc: Exception) -> None:
        LOGGER.error('Failed to execute query for integration %s: %s',
                     integration_name, exc)
        raise errors.DatabaseError('Error loading Integration', error=exc)

    async with application.postgres_connector(
            on_error=on_postgres_error) as conn:
        result = await conn.execute(
            'SELECT i.name, i.api_endpoint, i.api_secret, i.created_at,'
            '       i.created_by, i.last_modified_at, i.last_modified_by'
            '  FROM v1.integrations AS i'
            ' WHERE i.name = %(integration_name)s',
            {'integration_name': integration_name},
            metric_name='get-integration')
        if result.row_count:
            return Integration.model_validate(result.row)



from tornado import web

from . import handlers

URLS = [
    web.url(r'/components$',
            handlers.CollectionRequestHandler,
            name='components'),
    web.url(r'/components/(?P<package_url>.+)$',
            handlers.RecordRequestHandler,
            name='component'),
    web.url(r'^/projects/(?P<project_id>\d+)/components$',
            handlers.ProjectComponentsRequestHandler,
            name='project-components'),
]



from __future__ import annotations

import asyncio
import dataclasses
import logging
import typing

import imbi.opensearch
from imbi import errors, models
from imbi.clients import opensearch

if typing.TYPE_CHECKING:
    from imbi import app

LOGGER = logging.getLogger(__name__)

FACT_DATA_TYPES = {
    'boolean': 'boolean',
    'date': 'date',
    'decimal': 'float',
    'integer': 'integer',
    'string': 'text',
    'timestamp': 'date'
}

PROJECT = {
    'archived': {
        'type': 'boolean'
    },
    'components': {
        'type': 'nested',
        'properties': {
            'name': {
                'type': 'text',
                'fields': {
                    'keyword': {
                        'type': 'keyword',
                        'ignore_above': 256
                    }
                }
            },
            'version': {
                'type': 'text',
                'fields': {
                    'keyword': {
                        'type': 'keyword',
                        'ignore_above': 16
                    }
                }
            },
        }
    },
    'created_at': {
        'type': 'date'
    },
    'created_by': {
        'type': 'text'
    },
    'dependencies': {
        'type': 'integer'
    },
    'description': {
        'type': 'text'
    },
    'environments': {
        'type': 'text'
    },
    'gitlab_project_id': {
        'type': 'text'
    },
    'id': {
        'type': 'integer'
    },
    'last_modified_at': {
        'type': 'date'
    },
    'last_modified_by': {
        'type': 'text'
    },
    'name': {
        'type': 'text'
    },
    'namespace': {
        'type': 'text'
    },
    'namespace_slug': {
        'type': 'text'
    },
    'pagerduty_service_id': {
        'type': 'text'
    },
    'project_score': {
        'type': 'integer'
    },
    'project_type': {
        'type': 'text'
    },
    'project_type_slug': {
        'type': 'text'
    },
    'sentry_project_slug': {
        'type': 'text'
    },
    'slug': {
        'type': 'text'
    },
    'sonarqube_project_key': {
        'type': 'text'
    },
}


class ProjectIndex(imbi.opensearch.SearchIndex[models.Project]):
    """Class for interacting with the OpenSearch Project index"""

    INDEX = 'projects'

    def __init__(self, application) -> None:
        super().__init__(application, models.project)

    async def searchable_fields(self) -> list[dict]:
        fields = []
        index_mappings = await self._build_mappings()
        for key, defn in index_mappings.items():
            try:
                fields.append({
                    'name': key,
                    'type': defn['type'],
                    'count': 0,
                    'scripted': False,
                    'searchable': True,
                    'aggregatable': True,
                    'readFromDocValues': True
                })
            except KeyError:
                pass

        async with self.application.postgres_connector(
                on_error=self._on_postgres_error) as cursor:
            result = await cursor.execute(
                'SELECT name, data_type FROM v1.project_fact_types;')
            for row in result:
                fields.append({
                    'name': row['name'].lower().replace(' ', '_'),
                    'type': FACT_DATA_TYPES[row['data_type']],
                    'count': 0,
                    'scripted': False,
                    'searchable': True,
                    'aggregatable': True,
                    'readFromDocValues': True
                })
        return fields

    async def _build_mappings(self) -> dict:
        defn = dict(PROJECT)
        async with self.application.postgres_connector(
                on_error=self._on_postgres_error) as cursor:
            result = await cursor.execute(
                'SELECT name, data_type FROM v1.project_fact_types;')
            for row in result:
                defn[f'facts.{opensearch.sanitize_key(row["name"])}'] = {
                    'type': FACT_DATA_TYPES[row['data_type']]
                }
            result = await cursor.execute(
                'SELECT link_type FROM v1.project_link_types;')
            for row in result:
                defn[f'links.{opensearch.sanitize_key(row["link_type"])}'] = {
                    'type': 'text'
                }
            result = await cursor.execute('SELECT name FROM v1.environments;')
            for row in result:
                defn[f'urls.{opensearch.sanitize_key(row["name"])}'] = {
                    'type': 'text'
                }
            result = await cursor.execute('SELECT name FROM v1.integrations;')
            for row in result:
                defn[f'identifiers.{opensearch.sanitize_key(row["name"])}'] = {
                    'type': 'text'
                }
        return defn

    def _serialize_document(self, doc: models.Project) -> dict:
        output = dataclasses.asdict(doc)
        for key in ['namespace', 'project_type']:
            output[f'{key}_slug'] = output[key]['slug']
            output[key] = output[key]['name']
        return output

    @staticmethod
    def _on_postgres_error(metric_name: str, exc: Exception) -> None:
        LOGGER.error('Failed to execute query for collection %s: %s',
                     metric_name, exc)
        raise errors.DatabaseError(f'Error executing {metric_name}', error=exc)


class RequestHandlerMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.search_index = ProjectIndex(self.application)

    async def index_document(self, project_id: typing.Union[int, str]) -> None:
        value = await models.project(project_id, self.application)
        await self.search_index.index_document(value)


async def initialize(application: 'app.Application', build=False):
    index = ProjectIndex(application)
    LOGGER.info('Creating the project index')
    await index.create_index()
    LOGGER.info('Creating the project index mappings')
    await index.create_mapping()

    if build:
        async with application.postgres_connector() as conn:
            result = await conn.execute(
                'SELECT id FROM v1.projects ORDER BY id', {},
                'build-project-index')
            ids = [row['id'] for row in result]
            LOGGER.info('Queueing %i projects for indexing', len(ids))
            for project_id in ids:
                value = await models.project(project_id, application)
                await index.index_document(value)
            LOGGER.info('Queued %i projects for indexing', len(ids))
            while True:
                pending = \
                    await index.application.opensearch.documents_pending()
                if not pending:
                    break
                LOGGER.info('Waiting for %i projects to index', pending)
                await asyncio.sleep(2)
            LOGGER.info('Indexing complete')



import re

from . import base


class RequestHandler(base.AuthenticatedRequestHandler):

    NAME = 'dashboard'

    NAMESPACE_COUNTS = re.sub(
        r'\s+', ' ', """\
        SELECT a.namespace_id,
               b.name AS name,
               b.icon_class AS icon,
               count(a.*)
          FROM v1.projects AS a
          JOIN v1.namespaces AS b
            ON b.id = a.namespace_id
         WHERE archived IS FALSE
         GROUP BY a.namespace_id, b.name, b.icon_class
         ORDER BY b.name ASC""")

    PROJECT_COUNTS = re.sub(
        r'\s+', ' ', """\
        SELECT a.project_type_id,
               b.name AS name,
               b.plural_name AS plural,
               b.icon_class AS icon,
               b.slug AS slug,
               count(a.*)
          FROM v1.projects AS a
          JOIN v1.project_types AS b
            ON b.id = a.project_type_id
         WHERE archived IS FALSE
         GROUP BY a.project_type_id, b.name, b.plural_name, b.icon_class,
                  b.slug
         ORDER BY count(a.*) DESC""")

    async def get(self):
        namespaces = await self.postgres_execute(
            self.NAMESPACE_COUNTS, metric_name='dashboard-namespaces')
        project_types = await self.postgres_execute(
            self.PROJECT_COUNTS, metric_name='dashboard-project-types')
        self.send_response({
            'namespaces': namespaces.rows,
            'project_types': project_types.rows
        })



import re

from imbi import errors
from imbi.endpoints import base, projects


class CollectionRequestHandler(projects.ProjectAttributeCollectionMixin,
                               base.CollectionRequestHandler):
    NAME = 'project-identifiers'
    ITEM_NAME = 'project-identifier'
    ID_KEY = ['project_id', 'integration_name']
    FIELDS = ['external_id', 'integration_name']

    COLLECTION_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT external_id, integration_name, project_id, created_at,
               created_by, last_modified_at, last_modified_by
          FROM v1.project_identifiers
         WHERE project_id = %(project_id)s
         ORDER BY project_id, integration_name
        """)

    GET_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT external_id, integration_name, project_id,
               created_at, created_by, last_modified_at, last_modified_by
          FROM v1.project_identifiers
         WHERE project_id = %(project_id)s
           AND integration_name = %(integration_name)s
        """)

    POST_SQL = re.sub(
        r'\s+', ' ', """\
        INSERT INTO v1.project_identifiers
                    (external_id, integration_name, project_id,
                     created_at, created_by)
             VALUES (%(external_id)s, %(integration_name)s, %(project_id)s,
                     CURRENT_TIMESTAMP, %(username)s)
          RETURNING external_id, integration_name, project_id
        """)

    async def get(self, *args, **kwargs) -> None:
        result = await self.postgres_execute(
            'SELECT id FROM v1.projects WHERE id = %(project_id)s', kwargs,
            'get-project-id')
        if not result:
            raise errors.ItemNotFound()
        await super().get(*args, **kwargs)


class RecordRequestHandler(projects.ProjectAttributeCRUDMixin,
                           base.CRUDRequestHandler):
    NAME = 'project-identifier'
    ID_KEY = ['project_id', 'integration_name']

    GET_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT project_id, integration_name, external_id,
               created_at, created_by,
               last_modified_at, last_modified_by
          FROM v1.project_identifiers
         WHERE project_id = %(project_id)s
           AND integration_name = %(integration_name)s
        """)
    PATCH_SQL = re.sub(
        r'\s+', ' ', """\
        UPDATE v1.project_identifiers
           SET external_id = %(external_id)s,
               integration_name = %(integration_name)s,
               last_modified_by = %(username)s,
               last_modified_at = CURRENT_TIMESTAMP
         WHERE project_id = %(current_project_id)s
           AND project_id = %(project_id)s
           AND integration_name = %(current_integration_name)s
        RETURNING external_id
        """)
    DELETE_SQL = re.sub(
        r'\s+', ' ', """\
        DELETE FROM v1.project_identifiers
              WHERE project_id = %(project_id)s
                AND integration_name = %(integration_name)s
        """)



from __future__ import annotations

import asyncio
import copy
import dataclasses
import logging
import typing

from imbi import models, opensearch

if typing.TYPE_CHECKING:
    from imbi import app

LOGGER = logging.getLogger(__name__)

# Make sure that the field names match models.OperationsLog.SQL
OPS_LOG = {
    'id': {
        'type': 'integer'
    },
    'occurred_at': {
        'type': 'date'
    },
    'recorded_at': {
        'type': 'date'
    },
    'recorded_by': {
        'type': 'text'
    },
    'display_name': {
        'type': 'text',
    },
    'completed_at': {
        'type': 'date'
    },
    'project_id': {
        'type': 'integer'
    },
    'project_name': {
        'type': 'text'
    },
    'environment': {
        'type': 'text'
    },
    'change_type': {
        'type': 'text'
    },
    'description': {
        'type': 'text'
    },
    'link': {
        'type': 'text'
    },
    'notes': {
        'type': 'text'
    },
    'performed_by': {
        'type': 'text'
    },
    'ticket_slug': {
        'type': 'text'
    },
    'version': {
        'type': 'text'
    }
}


class OperationsLogIndex(opensearch.SearchIndex[models.OperationsLog]):
    """Class for interacting with the OpenSearch OperationsLog index"""

    INDEX = 'operations-log'

    def __init__(self, application) -> None:
        super().__init__(application, models.operations_log)

    async def _build_mappings(self):
        return copy.deepcopy(OPS_LOG)

    def _serialize_document(
            self, doc: models.OperationsLog) -> dict[str, typing.Any]:
        return dataclasses.asdict(doc)


class RequestHandlerMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.search_index = OperationsLogIndex(self.application)

    async def index_document(self, ops_log_id: typing.Union[int, str]) -> None:
        value = await models.operations_log(ops_log_id, self.application)
        await self.search_index.index_document(value)


async def initialize(application: 'app.Application', build=False):
    index = OperationsLogIndex(application)
    LOGGER.info('Creating the operations-log index')
    await index.create_index()
    LOGGER.info('Creating the operations-log index mappings')
    await index.create_mapping()

    if build:
        async with application.postgres_connector() as conn:
            result = await conn.execute(
                'SELECT id FROM v1.operations_log ORDER BY id', {},
                'build-operations-log-index')
            ids = [row['id'] for row in result]
            LOGGER.info('Queueing %i operations-logs for indexing', len(ids))
            for ops_log_id in ids:
                value = await models.operations_log(ops_log_id, application)
                await index.index_document(value)
            LOGGER.info('Queued %i opeations-logs for indexing', len(ids))
            while True:
                pending = \
                    await index.application.opensearch.documents_pending()
                if not pending:
                    break
                LOGGER.info('Waiting for %i projects to index', pending)
                await asyncio.sleep(2)
            LOGGER.info('Indexing complete')



import dataclasses
import typing

from imbi import automations, errors, models
from imbi.clients import github


async def _get_associated_namespaces(
    context: automations.AutomationContext,
    namespace_id: int,
) -> list[str]:
    """Get associated namespace slugs for a given namespace.

    Args:
        context: Automation context for database queries
        namespace_id: The namespace ID to look up associations for

    Returns:
        List of associated namespace slugs (may be empty if no associations
        exist)
    """
    result = await context.run_query(
        'SELECT DISTINCT n.slug '
        'FROM v1.namespaces n '
        'JOIN v1.namespace_associations na ON n.id = na.child_namespace_id '
        'WHERE na.parent_namespace_id = %(namespace_id)s', {
            'namespace_id': namespace_id,
        },
        metric_name='get-namespace-associations')

    return [row['slug'] for row in result.rows]


@dataclasses.dataclass
class GitHubContext:
    integration_name: str
    repository: github.Repository


async def create_repository(
    context: automations.AutomationContext,
    automation: models.Automation,
    project: models.Project,
) -> None:
    """Create a GitHub repository for an Imbi project

    This automation creates the GitHub repository, adds the ID of the
    new repository as a project identifier, and sets the GitHub link
    if we have one configured. The repository is created in the GitHub
    organization associated with the Project Type slug.

    The github repository information is saved in the context using the
    function object (eg, `create_repository`) as the key.

    @see https://docs.github.com/en/rest/repos/repos?apiVersion=2022-11-28
    #create-an-organization-repository
    """
    client = await github.create_client(context.application,
                                        automation.integration_name,
                                        context.user)

    org = project.project_type.github_org or project.project_type.slug

    settings = context.application.settings['automations']['github']
    default_attributes = settings.get('default_respository_attributes', {})

    repository = await client.create_repository(
        org=org,
        name=project.slug,
        project_id=project.id,
        namespace_slug=project.namespace.slug,
        description=project.description,
        **default_attributes)

    context[create_repository] = GitHubContext(
        integration_name=automation.integration_name, repository=repository)
    context.add_callback(delete_repository)
    context.note_progress(
        'created GitHub repository %s/%s (id=%s) for Imbi project %s', org,
        repository.name, repository.id, project.id)

    associated_namespaces = await _get_associated_namespaces(
        context, project.namespace.id)

    if associated_namespaces:
        teams_to_assign = [ns.lower() for ns in associated_namespaces]
    else:
        # No associations, use the namespace itself
        teams_to_assign = [project.namespace.slug.lower()]

    for team_slug in teams_to_assign:
        try:
            await client.add_team_to_repository(org, project.slug, team_slug)
            context.note_progress('added team %s to GitHub repository %s/%s',
                                  team_slug, org, project.slug)
        except Exception as e:
            context.note_progress(
                'warning: failed to add team %s to repository %s/%s: %s',
                team_slug, org, project.slug, str(e))

    await context.run_query(
        'INSERT INTO v1.project_identifiers'
        '            (external_id, integration_name, project_id,'
        '             created_at, created_by)'
        '     VALUES (%(github_id)s, %(integration_name)s, %(imbi_id)s,'
        '             CURRENT_TIMESTAMP, %(username)s)', {
            'github_id': repository.id,
            'imbi_id': project.id,
            'integration_name': automation.integration_name,
            'username': context.user.username,
        })
    context.note_progress(
        'registered GitHub identifier %s for Imbi project %s', repository.id,
        project.id)

    link_type_id = settings.get('project_link_type_id', None)
    if link_type_id is not None:
        await context.run_query(
            'INSERT INTO v1.project_links'
            '            (project_id, link_type_id, created_by, url)'
            '     VALUES (%(imbi_id)s, %(link_type_id)s, %(username)s,'
            '             %(repository_url)s)', {
                'imbi_id': project.id,
                'link_type_id': link_type_id,
                'repository_url': repository.html_url,
                'username': context.user.username,
            })
        context.note_progress('created GitHub link %r for Imbi project %s',
                              repository.html_url, project.id)


async def delete_repository(context: automations.AutomationContext,
                            _error: BaseException):
    """Compensating action that deletes a GitHub repository

    The repository information is expected to be in
    ``context[create_repository]``.
    """
    try:
        github_info = typing.cast(GitHubContext, context[create_repository])
    except KeyError:
        return

    try:
        client = await github.create_client(context.application,
                                            github_info.integration_name,
                                            context.user)
    except errors.ClientUnavailableError:
        pass
    else:
        org = github_info.repository.owner.login
        repo = github_info.repository.name
        await client.delete_repository(org, repo)



import re

from imbi.endpoints import base


class CollectionRequestHandler(base.CollectionRequestHandler):
    NAME = 'notification-filters'
    ITEM_NAME = 'notification-filter'
    ID_KEY = ['integration_name', 'notification_name', 'name']
    FIELDS = []

    COLLECTION_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT filter_name AS name, integration_name, notification_name,
               pattern, operation, value, action, created_at, created_by,
               last_modified_at, last_modified_by
          FROM v1.notification_filters
         WHERE integration_name = %(integration_name)s
           AND notification_name = %(notification_name)s
         ORDER BY name
        """)

    GET_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT filter_name AS name, integration_name, notification_name,
               pattern, operation, value, action, created_at, created_by,
               last_modified_at, last_modified_by
          FROM v1.notification_filters
         WHERE integration_name = %(integration_name)s
           AND notification_name = %(notification_name)s
           AND filter_name = %(name)s
        """)

    POST_SQL = re.sub(
        r'\s+', ' ', """\
    INSERT INTO v1.notification_filters
                (filter_name, integration_name, notification_name, pattern,
                 operation, value, action, created_at, created_by)
         VALUES (%(name)s, %(integration_name)s, %(notification_name)s,
                 %(pattern)s, %(operation)s, %(value)s, %(action)s,
                 CURRENT_TIMESTAMP, %(username)s)
      RETURNING filter_name AS name, integration_name, notification_name,
                pattern, operation, value, action
    """)

    @base.require_permission('admin')
    async def post(self, *args, **kwargs) -> None:
        await super().post(*args, **kwargs)


class RecordRequestHandler(base.CRUDRequestHandler):
    NAME = 'notification-filter'
    ID_KEY = ['integration_name', 'notification_name', 'name']

    GET_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT filter_name AS name, integration_name, notification_name,
               pattern, operation, value, action, created_at, created_by,
               last_modified_at, last_modified_by
          FROM v1.notification_filters
         WHERE integration_name = %(integration_name)s
           AND notification_name = %(notification_name)s
           AND filter_name = %(name)s
        """)

    PATCH_SQL = re.sub(
        r'\s+', ' ', """\
        UPDATE v1.notification_filters
           SET filter_name = %(name)s,
               integration_name = %(integration_name)s,
               notification_name = %(notification_name)s,
               pattern = %(pattern)s,
               operation = %(operation)s,
               value = %(value)s,
               action = %(action)s,
               last_modified_at = CURRENT_TIMESTAMP,
               last_modified_by = %(username)s
         WHERE filter_name = %(current_name)s
           AND integration_name = %(integration_name)s
           AND notification_name = %(notification_name)s
        """)

    DELETE_SQL = re.sub(
        r'\s+', ' ', """\
        DELETE FROM v1.notification_filters
         WHERE filter_name = %(name)s
           AND integration_name = %(integration_name)s
           AND notification_name = %(notification_name)s
        """)

    @base.require_permission('admin')
    async def delete(self, *args, **kwargs) -> None:
        await super().delete(*args, **kwargs)

    @base.require_permission('admin')
    async def patch(self, *args, **kwargs) -> None:
        await super().patch(*args, **kwargs)



from __future__ import annotations

import re
import typing

import jsonpatch
import jsonpointer
import psycopg2.errors
import psycopg2.errorcodes
import pydantic

from imbi import errors
from imbi.endpoints import base
from imbi.endpoints.components import models, scoring


class CollectionRequestHandler(base.PaginatedCollectionHandler):
    NAME = 'components'
    ITEM_NAME = 'component'
    ID_KEY = 'package_url'
    FIELDS = [
        'package_url', 'name', 'status', 'home_page', 'icon_class',
        'active_version'
    ]
    DEFAULTS = {
        'status': 'Active',
        'icon_class': 'fas fa-save',
        'active_version': None
    }

    COLLECTION_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT c.package_url, c."name", c.status, c.home_page, c.icon_class,
               c.active_version, COUNT(v.id) AS version_count,
               COUNT(p.project_id) AS project_count, c.created_at,
               c.created_by, c.last_modified_at, c.last_modified_by
          FROM v1.components AS c
          LEFT JOIN v1.component_versions AS v ON v.package_url = c.package_url
          LEFT JOIN v1.project_components AS p ON p.version_id = v.id
         WHERE c.package_url > %(starting_package)s
         GROUP BY c.package_url, c."name", c.status, c.home_page, c.icon_class,
                  c.active_version, c.created_at, c.created_by,
                  c.last_modified_at, c.last_modified_by
         ORDER BY c.package_url ASC
         LIMIT %(limit)s
        """)
    GET_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT package_url, "name", status, home_page, icon_class,
               active_version, created_at, created_by,
               last_modified_at, last_modified_by
          FROM v1.components
         WHERE package_url = %(package_url)s
        """)
    POST_SQL = re.sub(
        r'\s+', ' ', """\
        INSERT INTO v1.components
                    (package_url, "name", status, home_page,
                     active_version, icon_class, created_by)
             VALUES (%(package_url)s, %(name)s, %(status)s, %(home_page)s,
                     %(active_version)s, %(icon_class)s, %(username)s)
          RETURNING *
        """)

    def get_pagination_token_from_request(self) -> models.ComponentToken:
        return models.ComponentToken.from_request(self.request)

    @base.require_permission('admin')
    async def post(self, *args, **kwargs) -> None:
        await super().post(*args, **kwargs)


class RecordRequestHandler(base.CRUDRequestHandler):
    NAME = 'component'
    ID_KEY = 'package_url'
    GET_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT package_url, "name", status, home_page, icon_class,
               active_version, created_at, created_by,
               last_modified_at, last_modified_by
          FROM v1.components
         WHERE package_url = %(package_url)s
        """)
    DELETE_SQL = re.sub(
        r'\s+', ' ', """\
        DELETE FROM v1.components WHERE package_url = %(package_url)s
        """)
    PATCH_SQL = re.sub(
        r'\s+', ' ', """\
        UPDATE v1.components
           SET package_url = %(package_url)s,
               "name" = %(name)s,
               status = %(status)s,
               home_page = %(home_page)s,
               icon_class = %(icon_class)s,
               active_version = %(active_version)s,
               last_modified_at = CURRENT_TIMESTAMP,
               last_modified_by = %(username)s
         WHERE package_url = %(current_package_url)s
        """)

    @base.require_permission('admin')
    async def delete(self, *args, **kwargs):
        await super().delete(*args, **kwargs)

    @base.require_permission('admin')
    async def patch(self, *args, **kwargs):
        result = await self.postgres_execute(
            self.GET_SQL, {'package_url': kwargs['package_url']},
            f'get-{self.NAME}')
        if not result:
            raise errors.ItemNotFound(instance=self.request.uri)

        original = models.Component.model_validate(result.row)
        original_dict = original.model_dump()
        try:
            patch = jsonpatch.JsonPatch(self.get_request_body())
        except (jsonpatch.JsonPatchException,
                jsonpointer.JsonPointerException) as error:
            raise errors.BadJsonPatch(error)
        updated_dict = patch.apply(original_dict)
        if all(original_dict[k] == updated_dict[k] for k in updated_dict):
            self._add_self_link(self.request.path)
            self._add_link_header()
            self.set_status(304)
            return

        try:
            updated = models.Component.model_validate(updated_dict)
        except pydantic.ValidationError as error:
            all_errors = error.errors(include_context=False)
            raise errors.BadRequest(
                'failed to validate patched version: %s',
                str(error).replace('\n', ';'),
                title='Invalid Component generated by update',
                detail=all_errors[0]['msg'],
                validation_errors=all_errors) from None

        updated_dict.update({
            'current_package_url': original.package_url,
            'username': self._current_user.username,
        })
        result = await self.postgres_execute(self.PATCH_SQL, updated_dict,
                                             f'patch-{self.NAME}')
        if not result:
            raise errors.DatabaseError('No rows were returned from PATCH_SQL',
                                       title='failed to update record')

        # the following is used to propagate foreign key violations
        # to our code ... these can happen if a project is removed
        # between the time that we retrieve the affected projects
        # and when we upsert the project fact
        def on_postgres_error(metric_name: str,
                              exc: Exception) -> typing.Optional[Exception]:
            if isinstance(exc, psycopg2.errors.ForeignKeyViolation):
                raise exc
            else:
                return self.on_postgres_error(metric_name, exc)

        tags = {'key': 'bulk_update', 'endpoint': self.NAME}
        async with self.application.postgres_connector(
                on_postgres_error, self.on_postgres_timing) as connector:
            if self._version_update_required(original, updated):
                tags['operation'] = 'component_versions'
                async with self.application.stats.track_duration(tags):
                    await scoring.update_component_version_statuses(
                        updated, connector)
            if self._project_update_required(original, updated):
                tags['operation'] = 'component_score'
                async with self.application.stats.track_duration(tags):
                    await self._update_affected_projects(updated.package_url)

        await self._get({'package_url': updated.package_url})

    def _version_update_required(self, original: models.Component,
                                 updated: models.Component) -> bool:
        if original.status != updated.status:
            self.logger.debug('VERSION UPDATE REQUIRED: status=%s->%s',
                              original.status, updated.status)
            return True
        if (updated.status == models.ComponentStatus.ACTIVE
                and updated.active_version != original.active_version):
            self.logger.debug('VERSION UPDATE REQUIRED: active_version=%r->%r',
                              original.active_version, updated.active_version)
            return True
        self.logger.debug(
            'VERSION UPDATE NOT REQUIRED status=%s->%s active_version=%r->%r',
            original.status, updated.status, original.active_version,
            updated.active_version)
        return False

    def _project_update_required(self, original: models.Component,
                                 updated: models.Component) -> bool:
        # only update if we have a fact id to worry about
        config = self.settings['component_scoring']
        if config.get('project_fact_type_id') is not None:
            # status change will always affect component scores
            if updated.status != original.status:
                self.logger.debug('PROJECT UPDATE REQUIRED: status=%s->%s',
                                  original.status, updated.status)
                return True
            # if the final status is active, then we only care if the
            # active version has changed
            if (updated.status == models.ComponentStatus.ACTIVE
                    and updated.active_version != original.active_version):
                self.logger.debug(
                    'PROJECT UPDATE REQUIRED: active_version=%s->%s',
                    original.active_version, updated.active_version)
                return True
        self.logger.debug(
            'PROJECT UPDATE NOT REQUIRED: fact_type_id=%r status=%s->%s'
            ' active_version=%s->%s', config.get('project_fact_type_id'),
            original.status, updated.status, original.active_version,
            updated.active_version)
        return False

    async def _update_affected_projects(self, package_url: str) -> None:
        # the following is used to propagate foreign key violations
        # to our code ... these can happen if a project is removed
        # between the time that we retrieve the affected projects
        # and when we upsert the project fact
        def on_postgres_error(metric_name: str,
                              exc: Exception) -> typing.Optional[Exception]:
            if isinstance(exc, psycopg2.errors.ForeignKeyViolation):
                raise exc
            else:
                return self.on_postgres_error(metric_name, exc)

        result = await self.postgres_execute(
            'SELECT DISTINCT project_id, version_id'
            '  FROM v1.project_components'
            ' WHERE package_url = %(package_url)s',
            {'package_url': package_url},
            metric_name='retrieve-affected-projects')
        if result:
            version_ids = {row['version_id'] for row in result}
            self.logger.info('found %s versions affected by this change',
                             len(version_ids))

            project_ids = {row['project_id'] for row in result}
            self.logger.info('found %s projects affected by this change',
                             len(project_ids))

            async with self.application.postgres_connector(
                    on_postgres_error, self.on_postgres_timing) as connector:
                for project_id in project_ids:
                    try:
                        await scoring.update_component_score_for_project(
                            project_id, connector, self.application)
                    except psycopg2.errors.ForeignKeyViolation as exc:
                        self.logger.warning(
                            'failed to update component score for project'
                            ' id %s: %s', project_id, exc)


class ProjectComponentsRequestHandler(base.PaginatedCollectionHandler):
    COLLECTION_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT c.package_url, c."name", c.icon_class,
               c.active_version, v.version, v.status
          FROM v1.project_components AS p
          JOIN v1.component_versions AS v ON v.id = p.version_id
          JOIN v1.components AS c ON c.package_url = v.package_url
         WHERE c.package_url > %(starting_package)s
           AND p.project_id = %(project_id)s
         ORDER BY c.package_url ASC
        """)

    def _postprocess_item(self, item) -> None:
        """Add component links to our response items"""
        item['link'] = self.reverse_url('component', item['package_url'])

    def get_pagination_token_from_request(
            self, *, project_id: str) -> models.ProjectComponentsToken:
        return models.ProjectComponentsToken.from_request(
            self.request, project_id=project_id)



import re

from imbi.endpoints import base


class _RequestHandlerMixin:

    ID_KEY = 'id'
    FIELDS = ['id', 'role_arn', 'environment', 'namespace_id']

    GET_SQL = re.sub(
        r'\s+', ' ', """\
            SELECT id, role_arn, environment, namespace_id,
                   created_at, created_by
              FROM v1.aws_roles
             WHERE id = %(id)s""")


class CollectionRequestHandler(_RequestHandlerMixin,
                               base.CollectionRequestHandler):

    NAME = 'aws-roles'
    ITEM_NAME = 'aws-role'

    COLLECTION_SQL = re.sub(
        r'\s+', ' ', """ \
        SELECT id, role_arn, environment, namespace_id, created_at, created_by
          FROM v1.aws_roles
      ORDER BY id""")

    POST_SQL = re.sub(
        r'\s+', ' ', """\
    INSERT INTO v1.aws_roles
                (role_arn, environment, namespace_id, created_by)
         VALUES (%(role_arn)s, %(environment)s, %(namespace_id)s,
                 %(username)s)
      RETURNING id""")


class RecordRequestHandler(_RequestHandlerMixin, base.AdminCRUDRequestHandler):

    NAME = 'aws-role'

    DELETE_SQL = re.sub(
        r'\s+', ' ', """\
        DELETE FROM v1.aws_roles
         WHERE id = %(id)s""")

    PATCH_SQL = re.sub(
        r'\s+', ' ', """\
        UPDATE v1.aws_roles
           SET role_arn = %(role_arn)s,
               environment = %(environment)s,
               namespace_id = %(namespace_id)s
         WHERE id = %(id)s""")



import typing

import pydantic

from imbi import common


class Param(pydantic.BaseModel):
    type: typing.Literal['String', 'SecureString', 'StringList']
    value: str


class EnvironmentFieldMixin:
    @pydantic.computed_field
    @property
    def environment(self) -> str:
        return self.path.get_parts()[0]


class RemovePatch(pydantic.BaseModel, EnvironmentFieldMixin):
    op: typing.Literal['remove']
    path: common.JsonPointer

    @pydantic.field_validator('path', mode='after')
    def validate_path(cls, value: common.JsonPointer,
                      info: pydantic.ValidationInfo):
        path_parts = value.get_parts()
        if len(path_parts) > 1:
            raise ValueError(
                f'Invalid path: {value}. Can only remove an environment')

        environments = info.context.get('environments', [])
        environment = path_parts[0]
        if environment not in environments:
            raise ValueError(
                f'Invalid environment "{environment}", must be one of: '
                f'{environments}')

        return value


class ReplacePatch(pydantic.BaseModel, EnvironmentFieldMixin):
    op: typing.Literal['replace']
    path: common.JsonPointer
    value: str

    @pydantic.computed_field
    @property
    def type(self) -> str:
        return self.path.get_parts()[1]

    @pydantic.model_validator(mode='after')
    def validate_replace(self, info: pydantic.ValidationInfo):
        environments = info.context.get('environments', [])

        try:
            [environment, update_type] = self.path.get_parts()
        except ValueError:
            raise ValueError(
                f'Invalid path: {self.value}. Must replace type or value at '
                'an environment')

        if environment not in environments:
            raise ValueError(
                f'Invalid environment "{environment}", must be one of: '
                f'{environments}')

        if update_type not in {'type', 'value'}:
            raise ValueError(
                f'Invalid update for "{update_type}", must be either "type" '
                'or "value"')

        if update_type == 'type' and self.value not in {
                'String', 'SecureString', 'StringList'
        }:
            raise ValueError(
                'Invalid SSM type update: value must be one of "String", '
                f'"StringList", or "SecureString", got {self.value}')

        return self


class AddPatch(pydantic.BaseModel, EnvironmentFieldMixin):
    op: typing.Literal['add']
    path: common.JsonPointer
    value: Param

    @pydantic.field_validator('path')
    @classmethod
    def validate_path(cls, value: common.JsonPointer,
                      info: pydantic.ValidationInfo):
        path_parts = value.get_parts()
        if len(path_parts) > 1:
            raise ValueError(
                f'Invalid path: {value}. Can only add at an environment')

        environments = info.context.get('environments', [])
        environment = path_parts[0]
        if environment not in environments:
            raise ValueError(
                f'Invalid environment "{environment}", must be one of: '
                f'{environments}')

        return value


SSMPatch = typing.Annotated[typing.Union[RemovePatch, ReplacePatch, AddPatch],
                            pydantic.Field(..., discriminator='op')]


class SSMPatchBody(pydantic.RootModel):
    root: list[SSMPatch]

    def __iter__(self):
        return iter(self.root)

    def __getitem__(self, item):
        return self.root[item]



import uuid

from tests import base


class EncryptionTestCase(base.TestCase):
    def test_lifecycle(self):
        value = str(uuid.uuid4())
        encrypted = self._app.encrypt_value(value)
        self.assertEqual(self._app.decrypt_value(encrypted), value)


class PasswordHashingTestCase(base.TestCase):
    def test_that_password_hashes_include_algorithm(self):
        hashed = self._app.hash_password('my password')
        self.assertTrue(
            hashed.startswith(self._app.keychain.algorithm.name + ':'),
            'Hashed password is not prefixed by algorithm')



from tornado import web

from . import filters
from . import notifications
from . import processing
from . import rules

PREFIX = r'^/integrations/(?P<integration_name>[^/]+)/notifications'
SLUG = r'[\w_\-%\+]+'

URLS = [
    web.url(fr'{PREFIX}/?$',
            notifications.CollectionRequestHandler,
            name=notifications.CollectionRequestHandler.NAME),
    web.url(fr'{PREFIX}/(?P<name>{SLUG})$',
            notifications.RecordRequestHandler,
            name=notifications.RecordRequestHandler.NAME),
    web.url(fr'{PREFIX}/(?P<notification_name>{SLUG})/filters$',
            filters.CollectionRequestHandler,
            name=filters.CollectionRequestHandler.NAME),
    web.url(
        fr'{PREFIX}/(?P<notification_name>{SLUG})/filters/(?P<name>{SLUG})$',
        filters.RecordRequestHandler,
        name=filters.RecordRequestHandler.NAME),
    web.url(fr'{PREFIX}/(?P<notification_name>{SLUG})/rules$',
            rules.CollectionRequestHandler,
            name=rules.CollectionRequestHandler.NAME),
    web.url(
        fr'{PREFIX}/(?P<notification_name>{SLUG})/rules'
        r'/(?P<fact_type_id>\d+)$',
        rules.RecordRequestHandler,
        name=rules.RecordRequestHandler.NAME),
    web.url(
        fr'{PREFIX}/(?P<notification_name>{SLUG})'
        r'/(?P<action>delete|get|post|put)$',
        processing.ProcessingHandler,
        name='notification-processing'),
]



"""
System Reports

"""
from tornado import web

from . import (component_usage, namespace_kpis, namespace_shs_history,
               outdated_projects, system_shs_history)

URLS = [
    web.url(r'/reports/component-usage', component_usage.RequestHandler),
    web.url(r'/reports/namespace-kpis', namespace_kpis.RequestHandler),
    web.url(r'/reports/namespace-shs-history',
            namespace_shs_history.RequestHandler),
    web.url(r'/reports/outdated-projects', outdated_projects.RequestHandler),
    web.url(r'/reports/system-shs-history', system_shs_history.RequestHandler),
]



import base64
import http.client
import logging
import pathlib
import stat
import typing
import urllib.parse

import pydantic
import sprockets.mixins.http
import tornado.web
import yarl

from imbi import errors, oauth2, user, version
if typing.TYPE_CHECKING:
    from imbi import app


class Namespace(pydantic.BaseModel):
    id: int


class ProjectLinks(pydantic.BaseModel):
    self: str


class ProjectInfo(pydantic.BaseModel):
    id: int
    default_branch: str
    namespace: Namespace
    name_with_namespace: str
    web_url: str
    links: ProjectLinks = pydantic.Field(..., alias='_links')


PROJECT_DEFAULTS = {
    'auto_devops_enabled': False,
    'default_branch': 'main',
    'builds_access_level': 'enabled',
    'forking_access_level': 'enabled',
    'initialize_with_readme': False,
    'merge_requests_access_level': 'enabled',
    'only_allow_merge_if_pipeline_succeeds': True,
    'operations_access_level': 'disabled',
    'packages_enabled': False,
    'printing_merge_request_link_enabled': True,
    'remove_source_branch_after_merge': True,
    'snippets_access_level': 'enabled',
    'squash_option': 'default_off',
    'visibility': 'public',
}


class GitLabAPIFailure(errors.ApplicationError):
    def __init__(self, response: sprockets.mixins.http.HTTPResponse,
                 log_message: str, *log_args, **kwargs) -> None:
        status_code = response.code
        kwargs.setdefault('title', 'GitLab API Failure')
        kwargs['gitlab_response_body'] = response.body

        # A GitLab "unauthorized" response is not the same as an
        # Imbi "unauthorized". HTTP unauthorized always refers to
        # the user requested resource which is an Imbi resource.
        if status_code == http.HTTPStatus.UNAUTHORIZED:
            status_code = http.HTTPStatus.FORBIDDEN

        super().__init__(status_code, 'gitlab-api-failure', log_message,
                         *log_args, **kwargs)


async def create_client(
    application: 'app.Application',
    integration_name: str,
    current_user: user.User,
) -> '_GitLabClient':
    tokens = await current_user.fetch_integration_tokens(integration_name)
    if not tokens:
        raise errors.ClientUnavailableError(
            integration_name,
            f'no {integration_name!r} token for {current_user.username}')

    return _GitLabClient(current_user, tokens[0], application)


class _GitLabClient(sprockets.mixins.http.HTTPClientMixin):
    """API Client for GitLab.

    This client sends authenticated HTTP requests to the GitLab API.

    """
    def __init__(self, user: user.User, token: oauth2.IntegrationToken,
                 application: tornado.web.Application):
        super().__init__()
        settings = application.settings['automations']['gitlab']
        self.logger = logging.getLogger(__package__).getChild('GitLabClient')
        self.restrict_to_user = settings.get('restrict_to_user', False)
        self.user = user
        self.token = token
        self.headers = {
            'Accept': 'application/json',
            'Authorization': f'Bearer {self.token.access_token}',
        }
        self._user_info: typing.Union[dict, None] = None
        self._user_namespace: typing.Union[dict, None] = None

    async def api(self,
                  url: typing.Union[yarl.URL, str],
                  *,
                  method: str = 'GET',
                  **kwargs):
        """Make an authenticated request to the GitLab API."""
        if not isinstance(url, yarl.URL):
            url = yarl.URL(url)
        if not url.is_absolute():
            url = self.token.integration.api_endpoint / url.path.lstrip('/')
        request_headers = kwargs.setdefault('request_headers', {})
        request_headers.update(self.headers)
        if kwargs.get('body', None) is not None:
            request_headers['Content-Type'] = 'application/json'
            kwargs['content_type'] = 'application/json'
        kwargs['user_agent'] = f'imbi/{version} (GitLabClient)'
        self.logger.debug('%s %s', method, url)

        response = await super().http_fetch(str(url), method=method, **kwargs)
        if response.code == 400 and response.body:
            self.logger.warning('%s %s failed: status=%s response=%r', method,
                                url, response.code, response.body)
        if response.code == http.client.UNAUTHORIZED:  # maybe refresh
            if response.body.get('error', '') == 'invalid_token':
                await self._refresh_token()
                request_headers.update(self.headers)
                response = await super().http_fetch(str(url),
                                                    method=method,
                                                    **kwargs)
        return response

    async def fetch_all_pages(self, *path, **query) -> list[dict]:
        url = self.token.integration.api_endpoint
        for component in path:
            url /= str(component)
        if query:
            url = url.with_query(query)

        entries = []
        while url is not None:
            response = await self.api(url)
            if not response.ok:
                raise GitLabAPIFailure(response, 'GET %s failed: %s', url,
                                       response.code)

            entries.extend(response.body)
            for link in response.links:
                if link['rel'] == 'next':
                    url = yarl.URL(link['target'])
                    break
            else:
                url = None
        return entries

    async def fetch_group(self, *group_path) -> typing.Optional[dict]:
        slug = urllib.parse.quote('/'.join(group_path), safe='')
        url = self.token.integration.api_endpoint
        path = f'{url.path.rstrip("/")}/groups/{slug}'
        url = url.with_path(path, encoded=True).with_query(
            {'with_projects': 'false'})
        response = await self.api(url)
        if response.code == 404:
            self.logger.debug('Group not found: %s', url)
            return None
        if response.ok:
            return response.body
        raise GitLabAPIFailure(response, 'failed to fetch group %s: %s', slug,
                               response.code)

    async def fetch_project(self, project_id: int) \
            -> typing.Optional[ProjectInfo]:
        url = (self.token.integration.api_endpoint / 'projects' /
               str(project_id))
        response = await self.api(url)
        if response.code == 404:
            return None
        if response.ok:
            return ProjectInfo.parse_obj(response.body)
        raise GitLabAPIFailure(response, 'failed to fetch project %s: %s',
                               project_id, response.code)

    async def create_project(self, parent, project_name: str,
                             **attributes) -> ProjectInfo:
        if self.restrict_to_user:
            parent = await self.fetch_user_namespace()

        for name, value in PROJECT_DEFAULTS.items():
            attributes.setdefault(name, value)
        attributes.update({
            'name': project_name,
            'namespace_id': parent['id'],
        })
        response = await self.api('projects', method='POST', body=attributes)
        if response.ok:
            return ProjectInfo.parse_obj(response.body)
        raise GitLabAPIFailure(
            response, 'failed to create project %s in namespace %s: %s',
            project_name, parent['id'], response.code)

    async def delete_project(self, project: ProjectInfo) -> None:
        url = yarl.URL('/projects') / str(project.id)
        response = await self.api(url, method='DELETE')
        # intentionally swallow "Not Found" responses
        if not response.ok and response.code not in (http.HTTPStatus.NOT_FOUND,
                                                     http.HTTPStatus.GONE):
            raise GitLabAPIFailure(
                response,
                '%s failed to delete project %s',
                self.user.username,
                project.web_url,
                detail=f'Failed to delete {project.web_url}')

    async def commit_tree(self, project_info: ProjectInfo,
                          project_dir: pathlib.Path, commit_message: str):
        self.logger.info('creating commit for %s', project_info.id)

        files = [path for path in project_dir.glob('**/*') if path.is_file()]
        self.logger.debug('found %d files in %s', len(files), project_dir)

        actions = [{
            'action': 'create',
            'content': base64.b64encode(file.read_bytes()).decode('ascii'),
            'encoding': 'base64',
            'file_path': str(file.relative_to(project_dir)),
        } for file in files]
        actions.extend([{
            'action': 'chmod',
            'execute_filemode': True,
            'file_path': str(file.relative_to(project_dir)),
        } for file in files if file.stat().st_mode & stat.S_IXUSR])
        self.logger.debug('creating commit with %d actions', len(actions))

        project_url = yarl.URL(project_info.links.self)
        response = await self.api(
            project_url / 'repository' / 'commits',
            method='POST',
            body={
                'branch': project_info.default_branch,
                'commit_message': commit_message,
                'actions': actions,
            },
        )
        if response.ok:
            return response.body
        raise GitLabAPIFailure(response, 'failed to commit to %s: %s',
                               project_url, response.code)

    async def fetch_user_information(self):
        if self._user_info:
            return self._user_info
        response = await self.api(self.token.integration.api_endpoint / 'user')
        if response.ok:
            self._user_info = response.body
            return self._user_info
        raise GitLabAPIFailure(
            response, 'failed to retrieve GitLab user information: %s',
            response.code)

    async def fetch_user_namespace(self):
        if self._user_namespace:
            return self._user_namespace
        user_info = await self.fetch_user_information()
        response = await self.api(self.token.integration.api_endpoint /
                                  'namespaces' / user_info['username'])
        if response.ok:
            self._user_namespace = response.body
            return self._user_namespace
        raise GitLabAPIFailure(
            response,
            'failed to retrieve GitLab user namespace for %s: %s',
            user_info['username'],
            response.code,
            detail='failed to retrieve GitLab user namespace')

    async def _refresh_token(self) -> None:
        body = urllib.parse.urlencode({
            'grant_type': 'refresh_token',
            'refresh_token': self.token.refresh_token,
        })
        self.logger.info('refreshing token for %s (%s)', self.user.username,
                         self.token.external_id)
        self.logger.debug('refresh_request: %r', body)
        response = await super().http_fetch(
            str(self.token.integration.token_endpoint),
            method='POST',
            body=body,
            content_type='application/x-www-form-urlencoded',
            auth_username=self.token.integration.client_id,
            auth_password=self.token.integration.client_secret,
        )
        if response.ok:
            self.logger.debug('refresh response: %r', response.body)
            await self.token.integration.upsert_user_tokens(
                self.user.username, self.token.external_id,
                response.body['access_token'], response.body['refresh_token'],
                self.token.id_token)
            self.token.access_token = response.body['access_token']
            self.token.refresh_token = response.body['refresh_token']
            self.headers['Authorization'] = f'Bearer {self.token.access_token}'
        else:
            self.logger.error('failed to refresh token for %s: %r',
                              self.user.username, response.body)
            raise GitLabAPIFailure(response, 'token refresh failed: %s',
                                   response.code)



import opensearchpy

from imbi import errors
from imbi.endpoints import base


class RequestHandler(base.AuthenticatedRequestHandler):
    async def post(self, index: str):
        self.logger.debug('Request body: %r', self.get_request_body())
        try:
            result = await self.application.opensearch.client.search(
                index=index, body=self.get_request_body())
        except opensearchpy.RequestError as err:
            if err.status_code == 400:
                raise errors.BadRequest('OpenSearch request error',
                                        detail=err.info)
        self.send_response(result)
